/*
    Copyright (C) 1998-99 Paul Barton-Davis
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: pix.cc,v 1.4 2002/11/11 00:02:50 trutkin Exp $
*/

#include <algorithm>

#include <gdk/gdk.h>

#include <pbd/pathscanner.h>
#include <pbd/failed_constructor.h>

#include <gtkmm2ext/pix.h>

using namespace std;
using namespace Gtkmm2ext;

struct string_ptr_less {
    bool operator()(string *s1, string *s2) const {
	    return *s1 < *s2;
    }
};

Pix::PixCache *Pix::cache;

Pix::Pix (bool homog)
{
	_homegenous = homog;
}

Pix::Pix (vector<const char **> xpm_data, bool homog)
{
	if (xpm_data.size() == 0) {
		throw failed_constructor();
	}

	refcnt = 0;
	generated = false;
	max_pixwidth = 0;
	max_pixheight = 0;
	_homegenous = homog;

	data = xpm_data;
	pixmap_count = data.size();
	from_files = false;

}

Pix::Pix (const string &dirpath, const string &regexp, bool homog)
{
	PathScanner scanner;
	string_ptr_less cmp;

	refcnt = 0;
	generated = false;
	max_pixwidth = 0;
	max_pixheight = 0;
	_homegenous = homog;

	files = scanner (dirpath, regexp, false, true);

	sort (files->begin(), files->end(), cmp);

	if (files == 0) {
		return;
	}
	
	from_files = true;
	pixmap_count = files->size();
}

Pix::~Pix ()
{
	if (from_files) {
		for (vector<string*>::iterator i = files->begin(); i != files->end(); ++i) {
			delete *i;
		}
	}
}

void
Pix::generate (const Glib::RefPtr<const Gdk::Drawable> win, const Gdk::Color& transparent)
{
	if (generated) {
		return;
	}

	for (size_t i = 0; i < pixmap_count; i++) {
		int w, h;

		bitmaps.push_back (Glib::RefPtr<Gdk::Bitmap> ());

		if (from_files) {
			pixmaps.push_back (Gdk::Pixmap::create_from_xpm (win, bitmaps.back(), transparent, *((*files)[i])));
		} else {
			pixmaps.push_back (Gdk::Pixmap::create_from_xpm (win, bitmaps.back(), transparent, (const char* const*) data[i]));
		}
		
		pixmaps.back()->get_size (w, h);

		if (w > max_pixwidth) max_pixwidth = w;
		if (h > max_pixheight) max_pixheight = h;
	}

	generated = true;
}

Pix *
Gtkmm2ext::get_pix (string name, vector<const char **> xpm_data, bool homog)
{
	Pix *ret = 0;
	Pix::PixCache::iterator iter;
	pair<string, Pix *> newpair;

	if (Pix::cache == 0) {
		Pix::cache = new Pix::PixCache;
	}

	if ((iter = Pix::cache->find (name)) == Pix::cache->end()) {
		ret = new Pix (xpm_data, homog);
		if (ret->pixmap_count == 0) {
			delete ret;
			return 0;
		}
		newpair.first = name;
		newpair.second = ret;
		ret->cache_position = (Pix::cache->insert (newpair)).first;
		ret->refcnt++;
		return ret;
	} else {
		(*iter).second->refcnt++;
		return (*iter).second;
	}
}

Pix *
Gtkmm2ext::get_pix (const string &dirpath, const string &regexp, bool homog)
{
	Pix *ret = 0;
	Pix::PixCache::iterator iter;
	pair<string, Pix *> newpair;

	if (Pix::cache == 0) {
		Pix::cache = new Pix::PixCache;
	}

	if ((iter = Pix::cache->find (regexp)) == Pix::cache->end()) {
		ret = new Pix (dirpath, regexp, homog);
		if (ret->pixmap_count == 0) {
			delete ret;
			return 0;
		}
		newpair.first = regexp;
		newpair.second = ret;
		ret->cache_position = (Pix::cache->insert (newpair)).first;
		ret->refcnt++;
		return ret;
	} else {
		(*iter).second->refcnt++;
		return (*iter).second;
	}
}

void 
Gtkmm2ext::finish_pix (Pix *pix)
{
	pix->refcnt--;
	if (pix->refcnt == 0) {
		Pix::cache->erase (pix->cache_position);
		delete pix;
	}
}
