/*
    Copyright (C) 2000-2007 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: configuration.cc,v 1.2 2000/11/25 00:10:31 pbd Exp $
*/

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

#include <glib/gfileutils.h>
#include <glibmm/fileutils.h>
#include <glibmm/miscutils.h>

#include <pbd/failed_constructor.h>
#include <pbd/error.h>
#include <pbd/strsplit.h>

#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>

#include <midi++/manager.h>

using namespace std;
using namespace PBD;

#include "configuration.h"

Configuration::Configuration() 
{
	set_defaults ();
	load ();
}

void
Configuration::set_defaults ()
{
	use_real_time = false;
	real_time_priority = 10;
	n_steps = 16;
	n_sequences = 8;

	vector<Glib::ustring> v;

	v.push_back (CONFIG_DIR);
	v.push_back ("softwerk");
	v.push_back ("pixmaps");
	v.push_back ("knob.png");

	knob_image = Glib::build_filename (v);

	char *p = getenv ("SOFTWERK_UI_RC");

	if (p) {
		ui_rcfile = p;
	} else {
		v.clear ();
		v.push_back (CONFIG_DIR);
		v.push_back ("softwerk");
		v.push_back ("softwerk_ui.rc");
	}

	ui_rcfile = Glib::build_filename (v);

	pattern_directory = "/tmp";
	keybindings_path = "softwerk.bindings";
	ticks_per_beat = 16;
	default_midi_in = "default";
	default_midi_out = "default";
}

int
Configuration::load ()
{
	Glib::ustring rcfile;
	vector<Glib::ustring> v;
	
	/* load system configuration first */

	v.push_back (CONFIG_DIR);
	v.push_back ("softwerk");
	v.push_back ("softwerk.rc");

	rcfile = Glib::build_filename (v);

	if (Glib::file_test (rcfile, Glib::FILE_TEST_EXISTS)) {

		XMLTree tree;

		std::cerr << "Loading system configuration file " << rcfile << std::endl;

		if (!tree.read (rcfile.c_str())) {
			error << "SoftWerk: cannot read system configuration file "
			      << rcfile
			      << endmsg;
			return -1;
		}

		if (set_state (*tree.root())) {
			error <<"SoftWerk: system configuration file "
			      << rcfile
			      << " not loaded successfully."
			      << endmsg;
			return -1;
		}
	}

	/* load configuration file for user */
	v.clear ();
	
	v.push_back (getenv ("HOME"));
	v.push_back (".softwerkrc");

	rcfile = Glib::build_filename (v);

	if (Glib::file_test (rcfile, Glib::FILE_TEST_EXISTS)) {

		XMLTree tree;

		std::cerr << "Loading user configuration file " << rcfile << std::endl;

		if (!tree.read (rcfile)) {
			error << "SoftWerk: cannot read system configuration file "
			      << rcfile
			      << endmsg;
			return -1;
		}
		
		
		if (set_state (*tree.root())) {
			error <<"SoftWerk: system configuration file "
			      << rcfile
			      << " not loaded successfully."
			      << endmsg;
			return -1;
		}
	}

	return 0;
}

int
Configuration::set_state (const XMLNode& root)
{
	if (root.name() != "SoftWerk") {
		return -1;
	}

	XMLNodeList nlist = root.children();
	XMLNodeConstIterator niter;
	XMLNode *node;
	XMLProperty *prop;

	for (niter = nlist.begin(); niter != nlist.end(); ++niter) {

		node = *niter;

		if (node->name() == "MIDI-port") {

			try {
				MIDI::Port::Descriptor desc (*node);
				map<string,XMLNode>::iterator x;
				if ((x = midi_ports.find (desc.tag)) != midi_ports.end()) {
					midi_ports.erase (x);
				}
				midi_ports.insert (pair<string,XMLNode>(desc.tag,*node));
			}

			catch (failed_constructor& err) {
				warning << "ill-formed MIDI port specification in ardour rcfile (ignored)" << endmsg;
			}

		} else if (node->name() == "Config") {

			XMLNodeList option_list = node->children();
			XMLNodeConstIterator option_iter;
			XMLNode *option_node;

			string option_name;
			string option_value;

			for (option_iter = option_list.begin(); option_iter != option_list.end(); ++option_iter) {

				option_node = *option_iter;

				if (option_node->name() != "Option") {
					continue;
				}
				
				if ((prop = option_node->property ("name")) != 0) {
					option_name = prop->value();
				} else {
					return -1;
				}
				
				if ((prop = option_node->property ("value")) != 0) {
					option_value = prop->value();
				} else {
					return -1;
				}
				
				if (option_name == "initial-steps") {
					n_steps  = atoi (option_value.c_str());
				} else if (option_name == "initial-sequences") {
					n_sequences = atoi (option_value.c_str());
				} else if (option_name == "real-time") {
					use_real_time = (option_value == "yes");
				} else if (option_name == "real-time-priority") {
					real_time_priority = atoi (option_value.c_str());
				} else if (option_name == "ticks-per-beat") {
					ticks_per_beat = atoi (option_value.c_str());
				} else if (option_name == "pattern-directory") {
					pattern_directory = option_value;
				} else if (option_name == "default-midi-in") {
					default_midi_in = option_value;
				} else if (option_name == "default-midi-out") {
					default_midi_out = option_value;
				} else if (option_name == "ticks-per-beat") {
					ticks_per_beat = atoi (option_value.c_str());
				} else if (option_name == "knob-image") {
					knob_image = option_value;
				} else if (option_name == "ui-rc-file") {
					ui_rcfile = option_value;
				} else if (option_name == "keybindings-path") {
					keybindings_path = option_value;
				}
			}
		}
	}

	return 0;
}

string
Configuration::find_file (string name, string dir, string subdir)
{
	string path;
	char* envvar = getenv("SOFTWERK_PATH");

	/* 1st attempt: any directory in SOFTWERK_PATH */
	
	if (envvar != 0) {

		vector<string> split_path;
	
		split (envvar, split_path, ':');
		
		for (vector<string>::iterator i = split_path.begin(); i != split_path.end(); ++i) {
			path = *i;
			path += "/" + name;
			if (access (path.c_str(), R_OK) == 0) {
				// cerr << "Using file " << path << " found in ARDOUR_PATH." << endl;
				return path;
			}
		}
	}

	/* 2nd attempt: ~/.softwerk/ */

	path = Configuration::get_user_softwerk_path();
		
	if (subdir.length()) {
		path += subdir + "/";
	}
		
	path += name;
	if (access (path.c_str(), R_OK) == 0) {
		return path;
	}

	/* 3rd attempt: dir/... */
	
	path = dir;
	path += "/softwerk/";
	
	if (subdir.length()) {
		path += subdir + "/";
	}
	
	path += name;
	
	if (access (path.c_str(), R_OK) == 0) {
		return path;
	}

	return "";
}

string
Configuration::get_user_softwerk_path ()
{
	string path;
	char* envvar;
	
	if ((envvar = getenv ("HOME")) == 0 || strlen (envvar) == 0) {
		return "/";
	}
		
	path = envvar;
	path += "/.softwerk/";

	/* create it if necessary */

	if (g_mkdir_with_parents (path.c_str(), 0755)) {
		throw exception ();
	}

	return path;
}

string
Configuration::find_config_file (string name)
{
	const char* envvar;

	if ((envvar = getenv ("SOFTWERK_CONFIG_PATH")) == 0) {
		envvar = CONFIG_DIR;
	}

	return find_file (name, envvar);
}

