/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: main.cc,v 1.5 2000/11/25 00:10:31 pbd Exp $
*/

#include <string>

#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <midi++/port.h>
#include <midi++/port_request.h>
#include <midi++/manager.h>
#include <pbd/error.h>

#include <pbd/error.h>
#include <pbd/textreceiver.h>

#include "version.h"
#include "ui.h"
#include "configuration.h"
#include "sequencer.h"

TextReceiver   text_receiver ("SoftWerk");
Configuration *Config;

MIDI::Port *default_input_port = 0;
MIDI::Port *default_output_port = 0;

static Sequencer *sequencer;
static SequencerUI *ui;

using namespace std;
using namespace PBD;

static int 
setup_midi ()
{
	if (Config->midi_ports.empty()) {
		warning << "no MIDI ports specified. Adding a default one. This may not be what you want."
			<< endmsg;
		
		XMLNode new_port ("ignored");

		new_port.add_property ("mode", "duplex");
		new_port.add_property ("device", "softwerk");
		new_port.add_property ("type", "alsa/sequencer");
		new_port.add_property ("tag", "default");
			
		if (MIDI::Manager::instance()->add_port (new_port) == 0) {
			cerr << "darn, even that failed!\n";
			return -1;
		}
	}

	for (std::map<string,XMLNode>::iterator i = Config->midi_ports.begin(); i != Config->midi_ports.end(); ++i) {
		MIDI::Manager::instance()->add_port (i->second);
	}

	MIDI::Port* first;
	const MIDI::Manager::PortMap& ports = MIDI::Manager::instance()->get_midi_ports();

	if (ports.size() > 1) {

		first = ports.begin()->second;

		/* More than one port, so try using specific names for each port */

		if (Config->default_midi_in != "default") {
			default_input_port =  MIDI::Manager::instance()->port (Config->default_midi_in);
		} 

		if (Config->default_midi_out != "default") {
			default_output_port =  MIDI::Manager::instance()->port (Config->default_midi_out);
		} 

		/* If that didn't work, just use the first listed port */

		if (default_input_port == 0) {
			default_input_port = first;
		}

		if (default_output_port == 0) {
			default_output_port = first;
		}
		
	} else if (ports.size() == 1) {

		first = ports.begin()->second;

		/* Only one port described, so use it for both input & output */
		
		default_input_port = first;
		default_output_port = default_input_port;
	}

	if (default_input_port == 0) {
		warning << "no input port"
			<< endmsg;
		return -1;
	} 

	if (default_output_port == 0) {
		warning << "no output port"
			<< endmsg;
		return -1;
	}


	return 0;
}

int
main (int argc, char *argv[])
{
	bool use_rtc = false;

	Glib::thread_init ();

	while (1) {
		
		struct option long_options[] = {
		    { "use_rtc", 0, 0, 'r' },
		    { 0, 0, 0, 0 }
		};
		
		int c = getopt_long (argc, argv, "r", long_options, &optind);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 'r':
			use_rtc = true;
			break;
		default:
			exit (1);
		}
	}

	cout << "SoftWerk version " << "`Berlin'"
	     << ", Copyright (C) 1998-2008 Paul Davis" << endl
	     << "SoftWerk comes with ABSOLUTELY NO WARRANTY" << endl
	     << "This is free software, and you are welcome to redistribute it "
	     << endl
	     << "under certain conditions; see the file COPYING for details"
	     << endl;


	text_receiver.listen_to (error);
	text_receiver.listen_to (info);
	text_receiver.listen_to (fatal);
	text_receiver.listen_to (warning);

	Config = new Configuration ();

	if (setup_midi ()) {
		error << "cannot setup MIDI" << endmsg;
		return -1;
	}

	try { 
		ui = new SequencerUI (&argc, &argv, Config->ui_rcfile.c_str());
	}

	catch (...) {
		error << "could not create GUI" << endmsg;
		return -1;
	}


	try {
		sequencer = new Sequencer (Config->n_steps, Config->n_sequences, use_rtc);
	}

	catch (...) {
		error << "cannot create Sequencer" << endmsg;
		return -1;
	}
	
	if (ui->set_sequencer (*sequencer)) {
		error << "cannot use Sequencer" << endmsg;
		return -1;
	}

	ui->run (text_receiver);

	return 0;
}

