/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: patternfile.cc,v 1.2 2000/04/27 21:00:48 pbd Exp $
*/

#include <sstream>
#include <time.h>
#include <regex.h>
#include <pbd/error.h>
#include <pbd/pathscanner.h>

#include <midi++/port.h>

#include "patternfile.h"
#include "configuration.h"
#include "sequencer.h"
#include "trigger.h"

#include <pbd/failed_constructor.h>

using namespace PBD;

static const char *patternfile_regexp = "\\.swp$";

PatternFile::PatternFile (Sequencer& s)
	: sequencer (s)
{
	dir = Config->pattern_directory;
	edited = false;
}

PatternFile::PatternFile (Sequencer& s, const Glib::ustring& filename)
	: sequencer (s)
{
	dir = Config->pattern_directory;
	edited = false;

	set_name_from_path (filename);
	
	if (load()) {
		throw failed_constructor();
	}
}

void
PatternFile::set_name_from_path (const Glib::ustring& pathstr)
{
	Glib::ustring::size_type last_slash;
	Glib::ustring::size_type period;

	path = pathstr;
	
	last_slash = path.find_last_of ('/');
	
	if (last_slash != Glib::ustring::npos) {
		dir = path.substr (0, last_slash+1);
		_name = path.substr (last_slash+1);
	} else {
		dir = Config->pattern_directory;
		_name = path;
	}
	
	period = _name.find_last_of ('.');
	_name = _name.substr (0, period);
}

void
PatternFile::set_path_from_name ()
{
	path = dir;
	path += '/';
	path += _name;

	if (_name.find ('.') == string::npos) {
		path += ".swp";
	}
}

int
PatternFile::load ()
{
	int retval = -1;
	
	f.open (path.c_str());

	if (!f.good()) {
		error << "Cannot open pattern file \"" 
		      << path << "\""
		      << endmsg;
		goto done;
	}

	f.seekp (0, std::ios::end);

// XXX FIX ME	
//	if (f.tellp() == 0) {
//		/* empty file, don't parse it */
//		retval = 0;
//		goto done;
//	}

	f.seekp (0, std::ios::beg);

	if (parse()) {
		error << "Pattern file \""
		      << path
		      << "\" not parsed correctly" 
		      << endmsg;
		retval = -1;
		goto done;
	}

	retval = 0;

  done:
	f.close ();
	return retval;
}

int
PatternFile::save_trigger (Trigger& t)
{
	f << "! ";
	f << t;
	f << '\n';
	return 0;
}

int
PatternFile::save ()
{
	PatternMap::iterator i;
	time_t now;

	if (patterns.size() == 0) {
		edited = false;
		return 0;
	}

	/* we may not have a file yet - create on based on the current
	   name
	*/

	if (_name.length() == 0) {
		error << "PatternFile: you need to name this "
			"file before you can save it"
		      << endmsg;
		return -1;
	}
	
	set_path_from_name ();

	f.open (path.c_str(), std::ios::out|std::ios::trunc);

	if (!f) {
		error << "PatternFile: cannot save to \""
		      << path
		      << '"'
		      << endmsg;
		return -1;
	}

	time (&now);

	f << "SoftWerk ";
	f << ctime (&now);

	sequencer.foreach_trigger (*this, &PatternFile::save_trigger);

	for (i = patterns.begin(); i != patterns.end(); i++) {
		f << (*i).first << endl;
		f << (*i).second << endl;

		if (!f) {
			error << "write error while saving to pattern file \""
			      << path
			      << "'"
			      << endmsg;
			return -1;
		}
	}

	f.flush ();
	f.close ();

	edited = false;
	return 0;
}

int
PatternFile::parse ()
{
	char buf[4096];
	std::string foo;
	std::pair<Glib::ustring,std::string> newpair;

	/* File format:

	   first line: SoftWerk <some text><date><newline>
	   ! trigger-name .... \n
	   pattern name\n
	   pattern data\n

	   there may be any number of pattern name/data pairs
	*/

	f >> foo; /* "SoftWerk" */

	if (foo != "SoftWerk") {
		error << "pattern file \""
		      << path
		      << " is not a SoftWerk pattern file"
		      << endmsg;
		return -1;
	}

	f.getline (buf, sizeof (buf), '\n'); /* <some-text><date> */

	while (f) {
		f.getline (buf, sizeof (buf), '\n');

		if (f.gcount() == 0) {
			break;
		}

		if (buf[0] == '!') {
			Trigger* t = new Trigger;
			std::stringstream s;
			s << &buf[2];
			s >> *t;
			sequencer.add_trigger (*t);
			continue;
		}

		newpair.first = buf;
		f.getline (buf, sizeof (buf), '\n');

		if (f.gcount() == 0) {
			error << "format error in pattern file \""
			      << path
			      << " (pattern data missing for \""
			      << newpair.first
			      << "\")"
			      << endmsg;
			return -1;
		}

		newpair.second = buf;
		patterns.insert (newpair);
	}

	if (f.eof ()) {
		f.clear ();
		f.seekg (0, std::ios::beg);
	} else {
		error << "read error for pattern file \""
		      << path
		      << '"'
		      << endmsg;
		return (-1);
	}
	
	return (0);
}
	
PatternFile::~PatternFile ()
{
	close_file ();
}

void
PatternFile::reset ()
{
	edited = true;
	patterns.clear ();
}

std::string
PatternFile::pattern (const Glib::ustring& name)
{
	PatternMap::iterator i;

	if ((i = patterns.find (name)) == patterns.end()) {
		return 0;
	} else {
		return (*i).second;
	}

	return 0;
}

int
PatternFile::add_pattern (const Glib::ustring& name, const std::string& pattern)
{
	std::pair<Glib::ustring,std::string> newpair;
	std::pair<PatternMap::iterator,bool> result;

	newpair.first = name;
	newpair.second = pattern;
	
	result = patterns.insert (newpair);

	if (!result.second) {
		return -1;
	} 

	edited = true;
	return (0);
}

int
PatternFile::replace_pattern (const Glib::ustring& name, const std::string& pattern)
{
	PatternMap::iterator i;
	std::pair<Glib::ustring,std::string> newpair;

	if ((i = patterns.find (name)) == patterns.end()) {
		return -1;
	} 

	newpair.first = (*i).first;
	newpair.second = pattern;

	patterns.erase (i);
	patterns.insert (newpair);

	edited = true;
	return 0;
}

int
PatternFile::remove_pattern (const Glib::ustring& name)
{
	PatternMap::iterator i;

	if ((i = patterns.find (name)) == patterns.end()) {
		return -1;
	} 

	patterns.erase (i);

	edited = true;
	return (0);
}

int
PatternFile::rename_pattern (const Glib::ustring& oldname, const Glib::ustring& newname)
{
	PatternMap::iterator i;
	std::pair<Glib::ustring,std::string> newpair;

	if ((i = patterns.find (oldname)) == patterns.end()) {
		return -1;
	} 

	newpair.first = newname;
	newpair.second = (*i).second;
	
	patterns.erase (i);
	patterns.insert (newpair);

	edited = true;
	return 0;
}

void
PatternFile::close_file ()
{
	f.close ();
	path = "";
}

void
PatternFile::set_name (const Glib::ustring &str)
{
	if (str == _name) {
		return;
	}

	_name = str;
}
	
static bool
patternfile_filter (const string &str, void* ignored)
{
	static regex_t compiled_pattern;
	static bool first_time = true;
	static bool pattern_ok = false;

	if (first_time) {
		int err;
		char msg[1024];
		
		if ((err = 
		     regcomp (&compiled_pattern, patternfile_regexp,
			      REG_EXTENDED|REG_NOSUB|REG_ICASE|REG_NEWLINE))) {

			regerror (err, &compiled_pattern,
				  msg, sizeof (msg));

			error << "Cannot compile patternfile regexp for use ("
			      << msg << ')'
			      << endmsg;

			return false;
		}
		pattern_ok = true;
		first_time = false;
	}

	return pattern_ok && (regexec (&compiled_pattern,
				       str.c_str(), 0, 0, 0) == 0);
}

vector<string *> *
PatternFile::known_pattern_files (const string &searchpath)
{
	PathScanner scanner;

	return scanner (searchpath, patternfile_filter, false, true);
}
