#include <iostream>
#include <string>
#include <regex.h>

#include "patternfile.h"
#include "patternfile_ui.h"
#include "sequencer.h"

using namespace sigc;
using namespace PBD;
using namespace Gtk;

PatternFileUI::PatternFileUI (PatternFile &pf, Sequencer& s)
	: patternFile (pf),
	  sequencer (s),
	  addButton ("add current pattern"),
	  saveButton ("save"),
	  closeButton ("close"),
	  removeButton ("remove pattern"),
	  replaceButton ("replace pattern"),
	  renameButton ("rename pattern")
	
{
	std::string title;

	title = "SoftWerk: ";
	title += pf.name ();

	set_title (title);

	patternModel = ListStore::create (pattern_columns);
	patternView.set_model (patternModel);
	patternView.append_column ("Name", pattern_columns.name);
	patternView.set_headers_visible (false);
	patternView.set_reorderable (false);
	patternView.get_selection()->set_mode (SELECTION_SINGLE);
	patternView.set_size_request (200, 200);

	add (packer);

	packer.pack_start (lrBox, true, true, 10);

	lrBox.pack_start (patternView, true, true, 10);
	lrBox.pack_start (tbBox, true, true, 10);

	tbBox.set_spacing (10);
	tbBox.pack_start (modifyBox, false, false);
	tbBox.pack_start (nameBox, false, false);
	tbBox.pack_start (fileBox, false, false);
	
	fileBox.pack_start (closeButton, false, true, 10);
	fileBox.pack_end (saveButton, false, true, 10);
	
	modifyBox.set_spacing (10);
	modifyBox.pack_start (addButton, false, false);
	modifyBox.pack_start (renameButton, false, false);
	modifyBox.pack_start (replaceButton, false, false);
	modifyBox.pack_start (removeButton, false, false);

	nameLabel.set_text ("Pattern Set Name");
	nameEntry.set_text (patternFile.name());
	nameBox.pack_start (nameLabel, false, false, 10);
	nameBox.pack_start (nameEntry, true, false, 10);
	
	show_all ();

	refill ();

	patternView.get_selection()->signal_changed().connect(mem_fun (*this,&PatternFileUI::pattern_selected));
	patternView.signal_button_press_event().connect (mem_fun (*this, &PatternFileUI::button_press), false);

	closeButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::close));
	saveButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::save));
	addButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::add_pattern));
	removeButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::remove_pattern));
	replaceButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::replace_pattern));
	renameButton.signal_clicked().connect(mem_fun (*this,&PatternFileUI::rename_pattern));

	nameEntry.signal_changed().connect(mem_fun (*this,&PatternFileUI::rename_file));
}

PatternFileUI::~PatternFileUI ()
{
}

bool
PatternFileUI::button_press (GdkEventButton *ev)
{
	std::string pattern;
	std::stringstream pattern_stream;
	TreeIter iter;
	TreeModel::Path path;
	TreeViewColumn* column;
	int cellx;
	int celly;

	if (patternView.get_path_at_pos ((int)ev->x, (int)ev->y, path, column, cellx, celly)) {
		if ((iter = patternModel->get_iter (path))) {
			pattern = (*iter)[pattern_columns.pattern];
		}
	}

	if (pattern.empty()) {
		return false;
	}

	switch (ev->button) {
	case 1:
		if (ev->type == GDK_2BUTTON_PRESS) {
			pattern_stream << pattern;
			pattern_stream >> sequencer;
			return true;
		}
		return false;
		break;

	default:
		break; 
	}

	return false;
}	

void
PatternFileUI::pattern_selected ()
{
	selected_pattern.clear ();

	TreeView::Selection::ListHandle_Path rows = patternView.get_selection()->get_selected_rows ();

	if (rows.empty()) {
		return;
	}

	TreeView::Selection::ListHandle_Path::iterator i = rows.begin();
	TreeIter iter;
	
	if ((iter = patternModel->get_iter (*i))) {
		selected_pattern = (*iter)[pattern_columns.name];
	}
}
	
void
PatternFileUI::save ()
{
	patternFile.save ();
}

void
PatternFileUI::close ()
{
	if (patternFile.has_changed()) {
		/* XXX get an answer about saving it first */
		patternFile.save ();
	}

	hide_all ();
}

void
PatternFileUI::refill ()
{
	PatternFile::PatternMap::iterator i;

	patternModel->clear ();

	for (i = patternFile.patterns.begin(); i != patternFile.patterns.end(); ++i) {
		TreeModel::Row row = *(patternModel->append ());
		row[pattern_columns.name] = (*i).first;
		row[pattern_columns.pattern] = (*i).second;
	}
}

void
PatternFileUI::remove_pattern ()
{
	if (selected_pattern.empty () == 0) {
		error << "no pattern selected for removal" << endmsg;
		return;
	}

	if (patternFile.remove_pattern (selected_pattern)) {
		error << "pattern not found in pattern file!"
		      << endmsg;
	}

	refill ();
}

void
PatternFileUI::replace_pattern ()
{
	std::stringstream current_pattern;
	
	if (selected_pattern.empty()) {
		error << "no pattern selected for replacement" << endmsg;
		return;
	}

	current_pattern << sequencer;

	if (patternFile.replace_pattern (selected_pattern, current_pattern.str())) {
		error << "pattern not replaced!" 
		      << endmsg;
	}

	refill ();
	
}

void
PatternFileUI::rename_pattern ()
{
	if (selected_pattern.empty()) {
		error << "no pattern selected for removal" << endmsg;
		return;
	}

	Dialog d;
	Entry entry;
	Label label;
	Glib::ustring prompt;
	Glib::ustring name;

	prompt = "New name for \"";
	prompt += selected_pattern;
	prompt += "\":";

	label.set_text (prompt);

	d.get_vbox()->pack_start (label);
	d.get_vbox()->pack_start (entry);

	d.add_button (Stock::CANCEL, RESPONSE_CANCEL);
	d.add_button (Stock::OK, RESPONSE_ACCEPT);

	entry.signal_activate().connect (bind (mem_fun (d, &Dialog::response), RESPONSE_ACCEPT));

	d.set_transient_for (*this);
	d.show_all ();

	switch (d.run()) {
	case RESPONSE_ACCEPT:
		break;

	default:
		break;
	}

	name = entry.get_text ();

	if (!name.empty()) {
		if (patternFile.rename_pattern (selected_pattern, name)) {
			error << "pattern not renamed!" << endmsg;
			return;
		}
	}

	refill ();
}

void
PatternFileUI::add_pattern ()
{
	Dialog d;
	Entry entry;
	Label label;
	Glib::ustring prompt;
	Glib::ustring name;

	label.set_text ("New pattern name: ");

	d.get_vbox()->pack_start (label);
	d.get_vbox()->pack_start (entry);

	d.add_button (Stock::CANCEL, RESPONSE_CANCEL);
	d.add_button (Stock::OK, RESPONSE_ACCEPT);

	d.show_all ();
	d.set_transient_for (*this);

	entry.signal_activate().connect (bind (mem_fun (d, &Dialog::response), RESPONSE_ACCEPT));

	switch (d.run()) {
	case RESPONSE_ACCEPT:
		name = entry.get_text ();
		break;

	default:
		return;
		break;
	}

	name = entry.get_text ();

	if (name.empty()) {
		return;
	}
	std::stringstream current_pattern;
	
	current_pattern << sequencer;
	
	if (patternFile.add_pattern (name.c_str(), current_pattern.str())) {
		error << "pattern not added!"
		      << endmsg;
	} else {
		refill ();
	}
}

void
PatternFileUI::rename_file ()
{
	std::string name = nameEntry.get_text ();
	patternFile.set_name (name);
}
