/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: sequence.cc,v 1.7 2000/05/03 18:17:21 pbd Exp $
*/

#include <iostream>
#include <algorithm>
#include <cmath>

#include <sys/time.h>
#include <signal.h>
#include <cstring>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <midi++/manager.h>
#include <pbd/error.h>

#include "sequencer.h"
#include "step.h"
#include "trigger.h"
#include "configuration.h"

extern MIDI::Port* default_input_port;
extern MIDI::Port* default_output_port;

using namespace MIDI;
using namespace PBD;
using std::vector;
using std::min;

const char *Sequence::mode_strings[] = {
	"Absolute Pitch",
	"Relative Pitch",
	"Single Pitch (drum)",
	"MIDI bytes",
	"Recorded MIDI",
	"Gate Timing",
	"Note Length",
	"On Velocity",
	"Off Velocity",
	"Absolute Chord",
	"Relative Chord",
	"Program Change",
	"Controller",
	"Pitch Bend",
	"Polytouch",
	"Aftertouch",
	0
};

const char *Sequence::direction_strings[] = {
	"Forwards",
	"Backwards",
	"End-to-end",
	"Random",
	0
};


Sequence::Sequence (size_t rowid, size_t nsteps, Sequencer &seq)
	: _sequencer (seq)

{
	/* get whatever the default I/O ports are for MIDI */

	input_port = default_input_port;
	output_port = default_output_port;
	
	_id = rowid;

	for (size_t i = 0; i < nsteps; i++) {
		new_step ();
	}

	reset ();
}

Sequence::~Sequence ()
{
	if (output_port) {
		output_port->channel (_output_channel)->all_notes_off ();
	}
}

Step *
Sequence::new_step ()
{
	Glib::Mutex::Lock lm (_lock);
	Step *step;

	step = new Step (steps.size(), *this);	
	steps.push_back (step);

	return step;
}

void
Sequence::reset ()
{
	waiting_for_trigger = false;
	have_trigger = false;
	_trigger_mode = off_trigger;
	fired_cnt = 0;
	next_gate = 0; /* starts immediately */
	base_note = DEFAULT_BASE_NOTE;
	gate_time = Config->ticks_per_beat;
	note_time = Config->ticks_per_beat - 1;
	_current_step = 0;    /* make sure initial value is sane */
	set_current_step (0);
	_controller_id = 0;
	_input_channel = 0;
	_output_channel = 0;

	/* set_mode() is too heavyweight for this */

	_mode = relative_pitch;
	_modebits = (Sequence::note|Sequence::sound|Sequence::relative);

	_direction = forwards;
	e2emode = e2e_forward;
	controlled_rows = 0; /* control no rows */
	_volume = 64;
	_muted = false;
	_skip = false;
	_off = true;

	_first_step = 0;
	_start_step = 0;
	_last_step = steps.size() - 1;

	steps[_first_step]->set_start (true);
	steps[_last_step]->set_end (true);

	maybe_listen_to_note_on ();
}

void
Sequence::check (unsigned int now)

{
	vector<Step *>::iterator s;

	if (!makes_sound()) {
		return;
	}
    
	for (s = steps.begin(); s != steps.end(); s++) {
		(*s)->check (now);
	}
}

void
Sequence::advance ()

{
	Direction dir;
	size_t skip_cnt = 0;
	int cs;

	if (_off  || is_chord() || waiting_for_trigger) {
		return;
	}
    
	dir = _direction;

  again:
	cs = _current_step;

	switch (dir) {
	case forwards:
		if (++cs > (int) _last_step) { /* @end, going forwards */

			if (_direction == forwards) {
				reset_trigger ();
			}
	      
			if (_direction == end_to_end) {
				e2emode = e2e_backward;
				set_current_step  
					(std::max (_last_step - 1, _first_step));
			} else {
				set_current_step (_first_step);
			} 
		} else {
			set_current_step (cs);
		}

		break;
	  
	case backwards:
		if (--cs < (int) _first_step) { /* @start, going backwards */
			bool was_triggered;

			was_triggered = fired_cnt;

			if ((_direction == backwards || 
			     e2emode == e2e_backward)) { 
			    reset_trigger ();
			}

			if (_direction == end_to_end) {
				e2emode = e2e_forward;

				if (was_triggered) {
					set_current_step (_first_step);
				} else {
					set_current_step 
						(std::min (_first_step + 1, 
							   _last_step));
				}
			} else {
				set_current_step (_last_step);
			}

		} else {
			set_current_step (cs);
		}
		break;
	  
	case end_to_end:
		if (e2emode == e2e_forward) {
			dir = forwards;
			goto again;
		} else {
			dir = backwards;
			goto again;
		}
		break;
	  
	case rd_random:
		while ((cs = rand () % (_last_step + 1)) < (int) _first_step);
		set_current_step (cs);
		break;
	}

	if (_skip && !steps[_current_step]->is_enabled()) {
		if (++skip_cnt != active_steps()) {
			goto again;
		}
	}

}

void
Sequence::reset_trigger ()

{
	if (!have_trigger) {
		return;
	}

	if (_trigger_mode == off_trigger) {

		fired_cnt = 0;
		waiting_for_trigger = false;

	} else if (_trigger_mode == single) {
		/* one-shot trigger mode: wait for next
		   firing.
		*/
		
		fired_cnt = 0;
		waiting_for_trigger = true;
	} else {
		/* multi-trigger mode: reduce count of
		   firings by one, and if there are still
		   some waiting, then allow us to keep
		   running.
		*/

		if (fired_cnt > 0) {
			fired_cnt--;
		}
		if (fired_cnt == 0) {
			waiting_for_trigger = true;
		}
	}
}

void
Sequence::note_on (byte note)

{
	output_port->channel (_output_channel)->note_on (note, on_velocity ());
}

void
Sequence::note_off (byte note)

{
	output_port->channel (_output_channel)->note_off (note, off_velocity ());
}

byte
Sequence::on_velocity ()

{
	int x;

	if ((x = _sequencer.sumover (Sequence::note_on_velocity, _id)) < 0) {
		return _volume;
	} else {
		return std::min ((int) _volume, x);
	}
}

byte
Sequence::off_velocity ()

{
	int x;

	if ((x = _sequencer.sumover (Sequence::note_off_velocity, _id)) < 0) {
		return DEFAULT_OFF_VELOCITY;
	} else {
		return (byte) x;
	}
}

void
Sequence::play_chord (int onoff, ticks_t when)

{
	byte *msg;
	byte vel;
	int m;
	ticks_t dur;
	vector<Step *>::iterator s;

	msg = new byte [steps.size() * 3];
	vel = onoff ? on_velocity () : off_velocity ();
	dur = note_duration();

	for (s = steps.begin(), m = 0; s != steps.end(); s++) {

		if (!(*s)->is_enabled()) {
			continue;
		}

		msg[m++] = ((onoff ? MIDI::on : MIDI::off) << 4) |
			(_output_channel & 0xf);
		msg[m++] = (*s)->val ();
		msg[m++] = vel;

		(*s)->play (when, dur);
	}

	if (m) {
		midimsg (msg, m);
	}

	/* if fired by a trigger, say we've fired */

	reset_trigger ();
}

void
Sequence::on_notes_off ()

{
	vector<Step *>::iterator s;

	if (!makes_sound()) {
		return;
	}

	for (s = steps.begin(); s != steps.end(); s++) {
		(*s)->check (~0); /* force notes off */
	}
}


void
Sequence::set_mode (int m)
{
	bool preserve = false;
	{
		Glib::Mutex::Lock lm (_lock);
		
		if (makes_sound()) {
			on_notes_off ();
			preserve = true;
		} 	    
		
		_mode = (Sequence::Mode) m;
		_modebits = 0;
		
		switch (_mode) {
		case Sequence::absolute_pitch:
			_modebits = (Sequence::note|Sequence::sound);
			break;
		case Sequence::relative_pitch:
			_modebits = (Sequence::note|Sequence::sound|Sequence::relative);
			break;
		case Sequence::single_pitch:
			_modebits = (Sequence::note|Sequence::sound);
			break;
		case Sequence::recorded_midibytes:
		case Sequence::midibytes:
			_modebits = (Sequence::midi);
			break;
		case Sequence::gate_timing:
			_modebits = (Sequence::parameter);
			break;
		case Sequence::note_length:
			_modebits = (Sequence::parameter);
			break;
		case Sequence::note_on_velocity:
			_modebits = (Sequence::parameter);
			break;
		case Sequence::note_off_velocity:
			_modebits = (Sequence::parameter);
			break;
		case Sequence::absolute_chord:
			_modebits = (Sequence::chord|Sequence::sound);
			break;
		case Sequence::relative_chord:
			_modebits = (Sequence::chord|Sequence::sound|Sequence::relative);
			break;
		case Sequence::program_change:
			_modebits = (Sequence::channel);
			break;
		case Sequence::controller:
			_modebits = (Sequence::channel|Sequence::control_msg);
			break;
		case Sequence::pitch_bend:
			_modebits = (Sequence::channel);
			break;
		case Sequence::polytouch:
			_modebits = (Sequence::channel);
			break;
		case Sequence::aftertouch:
			_modebits = (Sequence::channel);
			break;
		default:
			fatal << "Sequence::set_bits: this can't happen" << endmsg;

		}	

		set_current_step (_first_step);
		set_default_step_values (preserve);
		maybe_listen_to_note_on ();

		/* XXX what should we do about an existing trigger */
	
	} /* end of atomic scope */

	/* tell anyone who cares that we just changed modes */

	mode_changed ();

	if (_mode == recorded_midibytes) {
		start_midi_grab ();
	}
}

void
Sequence::set_default_step_values (bool preserve)

{
	vector<Step *>::iterator s;
	
	if (makes_sound()) {

		if (!preserve) {
			set_base_note (DEFAULT_BASE_NOTE, 1); 
			/* set_base_note (random() % 127, 1);*/
		}

		set_steps_on (true);

	} else if (is_raw_midi()) {

		/* set MIDI messages to default: empty  */
	
		for (s = steps.begin(); s != steps.end(); s++) {

			/* need to force compiler's hand here due
			   to ambiguity in the overloading of
			   Step::set ();
			*/

			(*s)->set ((MIDI::byte *) 0, (size_t) 0);
		}
	
	}

	if (is_chord() || is_channel_msg()) {

		/* default chord/channelmsg state is no steps enabled */

		for (s = steps.begin(); s != steps.end(); s++) {
			(*s)->disable();

			if (is_channel_msg()) {
				(*s)->set (0);
			}
		}
	
	}
}

void
Sequence::set_last_step (size_t new_step)
{
	size_t old_step;

	{
		Glib::Mutex::Lock lm (_lock);

		new_step = abs (new_step % steps.size());
		
		steps[_last_step]->set_end (false);
		steps[new_step]->set_end (true);
		
		old_step = _last_step;
		_last_step = new_step;
		
		if (_last_step < _current_step) {
			set_current_step (_last_step);
		}
	}

	if (old_step > new_step && is_note()) {
		while (old_step != new_step) {
			note_off (steps[old_step]->val());
			old_step--;
		}
	}
}

void 
Sequence::set_first_step (size_t new_step)
{
	size_t old_step;

	{
		Glib::Mutex::Lock lm (_lock);
		
		new_step = abs (new_step % steps.size());
		
		steps[_first_step]->set_start (false);
		steps[new_step]->set_start (true);
		
		old_step = _first_step;
		_first_step = new_step;
		
		if (_first_step > _current_step) {
			set_current_step (_first_step);
		}
	}

	if (old_step < new_step && is_note()) {
		while (old_step != new_step) {
			note_off (steps[old_step]->val ());
			old_step++;
		}
	}
}

void 
Sequence::set_start_step (size_t new_step)
{
	Glib::Mutex::Lock lm (_lock);
	_set_start_step (new_step);
}

void 
Sequence::_set_start_step (size_t new_step)
{
	new_step = abs (new_step % steps.size());
	_start_step = new_step;
}

void
Sequence::set_direction (Direction dir)
{
	Glib::Mutex::Lock lm (_lock);
	_set_direction (dir);
}

void
Sequence::_set_direction (Direction dir)
{
	_direction = dir;

	if (dir == Sequence::end_to_end) {
		e2emode = Sequence::e2e_forward;

	} else if (is_note() && have_trigger) {
		if (dir == Sequence::backwards) {
			set_current_step (_last_step);

		} else if (dir == Sequence::forwards ||
			   dir == Sequence::end_to_end)  {
			set_current_step (_first_step);

		}
	}

}
 
void
Sequence::set_next_gate (ticks_t t)
{
	Glib::Mutex::Lock lm (_lock);
	_set_next_gate (t);
}

void
Sequence::_set_next_gate (ticks_t t)
{
	next_gate = t;
}

void
Sequence::set_all_steps (int note)
{
	Glib::Mutex::Lock lm (_lock);
	vector<Step *>::iterator s;

	for (s = steps.begin(); s != steps.end(); s++) {
		(*s)->set (note, true);
	}
}

void
Sequence::set_base_note (int note, bool absolute_note)
{
	Glib::Mutex::Lock lm (_lock);
	_set_base_note (note, absolute_note);
}

void
Sequence::_set_base_note (int note, bool absolute_note)
{
	int delta;
	vector<Step *>::iterator s;

	if (!is_relative()) {
		return;
	}
    
	on_notes_off ();
    
	delta = -(base_note - note);

	base_note = (absolute_note ? note : base_note + delta);

	for (s = steps.begin(); s != steps.end(); s++) {
		(*s)->set (absolute_note ? 
			   note : ((*s)->val() + delta) & 0x7F);
	}

}    

void
Sequence::set_read_port (MIDI::Port *p)
{
	Glib::Mutex::Lock lm (_lock);
	_set_read_port (p);
}

void
Sequence::_set_read_port (MIDI::Port *p)
{
	input_port = p;
	maybe_listen_to_note_on ();
	input_midi_changed ();
}
 
void
Sequence::set_write_port (MIDI::Port *p)
{
	Glib::Mutex::Lock lm (_lock);
	_set_write_port (p);
}

void
Sequence::_set_write_port (MIDI::Port *p)
{
	on_notes_off ();
	output_port = p;
	output_midi_changed ();
}

void
Sequence::set_input_channel (channel_t ch)
{
	Glib::Mutex::Lock lm (_lock);

	_set_input_channel (ch);
}

void
Sequence::_set_input_channel (channel_t ch)
{
	_input_channel = ch;
	maybe_listen_to_note_on ();
	input_midi_changed ();
}

void
Sequence::set_output_channel (channel_t ch)
{
	Glib::Mutex::Lock lm (_lock);

	_set_output_channel (ch);
}

void
Sequence::_set_output_channel (channel_t ch)
{
	on_notes_off ();
	_output_channel = ch;
	output_midi_changed ();
}

void
Sequence::maybe_listen_to_note_on ()
{
	note_on_connection.disconnect ();

	if (is_relative ()) {
		note_on_connection = 
			input_port->input()->
			    channel_note_on[(int)_input_channel].connect 
			         (mem_fun (*this, &Sequence::handle_note_on));
	} 
}

void
Sequence::handle_note_on (MIDI::Parser &p, MIDI::EventTwoBytes *ev)
{
	if (_off) {
		return;
	}

	set_base_note (ev->note_number, false);
}
   
ticks_t
Sequence::gate_interval ()

{
	int x;

	if ((x = _sequencer.sumover (Sequence::gate_timing, _id)) < 0) {
		return gate_time;
	} else {
		return (unsigned int) x;
	}
}

ticks_t
Sequence::note_duration ()

{
	int x;

	if ((x = _sequencer.sumover (Sequence::note_length, _id)) < 0) {
		return note_time;
	} else {
		return (unsigned int) x;
	}
}

void
Sequence::dump (std::ostream &str)
{
	vector<Step *>::iterator i;

	str << _id << ' '
	    << steps.size() << ' '
	    << _last_step << ' '
	    << _first_step << ' '
	    << _current_step << ' '
	    << base_note << ' '
	    << '"' << output_port->name() << "\" "
	    << (int) _output_channel << ' '
	    << '"' << input_port->name() << "\" "
	    << (int) _input_channel << ' '
	    << (int) _controller_id << ' '
	    << gate_time << ' '
	    << note_time << ' '
	    << controlled_rows << ' '
	    << (int) _volume << ' '
	    << (int) _mode << ' '
	    << (int) _modebits << ' '
	    << (int) e2emode << ' '
	    << (int) _direction << ' '
	    << (_skip ? 1 : 0) << ' '
	    << (_off ? 1 : 0) << ' '
	    << '"' << trigger_name << '"' << ' ';

	for (i = steps.begin(); i != steps.end(); i++) {
		str << **i << ' ';
	}
}

void
Sequence::load (std::istream &str)

{
	int foo;
	size_t nsteps;
	vector<Step *>::iterator i;
	char buf[64];
	char c;
	int fstep, lstep, cstep;
	string inport_name;
	string outport_name;
	MIDI::Port *port;

	reset();

	str >> _id;
	str >> nsteps;
	str >> lstep;
	str >> fstep;
	str >> cstep;
	str >> base_note;
	str >> c; /* eat " */
	str.getline (buf, sizeof (buf), '"');
	outport_name = buf;
	str >> foo;  _output_channel = (channel_t) foo;
	str >> c; /* eat " */
	str.getline (buf, sizeof (buf), '"');
	inport_name = buf;
	str >> foo;  _input_channel = (channel_t) foo;
	str >> foo; _controller_id = (byte) foo;
	str >> gate_time;
	str >> note_time;
	str >> controlled_rows;
	str >> foo; _volume = (byte) foo;
	str >> foo; _mode = (Mode) foo;
	str >> foo; _modebits = (ModeBits) foo;
	str >> foo; e2emode = (E2EMode) foo;
	str >> foo; _direction = (Direction) foo;
	str >> foo; _skip = foo;
	str >> foo; _off = foo;
	str >> c; /* eat " */
	str.getline (buf, sizeof (buf), '"');
	trigger_name = buf;
	
	while (steps.size () < nsteps) {
		new_step ();
	}

	for (i = steps.begin(); nsteps--; i++) {
		str >> **i;
	}

	/* set up MIDI ports */

	if ((port = MIDI::Manager::instance()->port (inport_name)) == 0){
		error << "MIDI port \"" 
		      << inport_name 
		      << "\" is unknown - using default port"
		      << endmsg;
		set_read_port (MIDI::Manager::instance()->inputPort);
	} else {
		set_read_port (port);
	}

	if ((port = MIDI::Manager::instance()->port (outport_name)) == 0){
		error << "MIDI port \"" 
		      << outport_name 
		      << "\" is unknown - using default port"
		      << endmsg;
		set_read_port (MIDI::Manager::instance()->outputPort);
	} else {
		set_write_port (port);
	}

	/* get trigger */

	if (trigger_name.length()) {
		Trigger *t;

		if ((t = _sequencer.trigger (trigger_name)) == 0) {
			error << "No such trigger \""
			      << trigger_name
			      << '"'
			      << endmsg;
		}

		use_trigger (*t);
	}

	set_first_step (fstep);
	set_last_step (lstep);
	set_current_step (cstep);
}

void
Sequence::get_next_wakeup (ticks_t& wakeup)
{
	if (_off) {
		return;
	}

	/* note: non-sound sequences have no "off" to take care of, so we don't
	   need to check steps to see if they have business to take care of
	   at a time other than the next gate time.
	*/
	if (makes_sound()) {
		for (vector<Step*>::iterator s = steps.begin(); s != steps.end(); ++s) {
			ticks_t t = (*s)->get_next_wakeup ();
			wakeup = min (wakeup, t);
		}
	}
	if (next_gate) {
		wakeup = min (next_gate, wakeup);
	}
}

void
Sequence::execute (size_t current_tick)

{
	if (!needs_attention (current_tick) || waiting_for_trigger) {
		return;
	}
	
	if (!makes_sound() || !silent()) {
		if (is_chord()) {
			play_chord (1, current_tick);
		} else {
			steps[_current_step]->execute (current_tick);
		}
	}

	_set_next_gate (current_tick + gate_interval ());
}

void
Sequence::foreach_step (void (*func)(Step &,void *), void *arg)

{
	vector<Step *>::iterator s;

	for (s = steps.begin(); s != steps.end(); s++) {
		func (**s, arg);
	}
}

void
Sequence::set_current_step (size_t s)

{
	steps[_current_step]->set_current (false);
	_current_step = s;
	steps[_current_step]->set_current (true);
}

int
Sequence::maximum_step_value ()

{
	if (is_raw_midi()) {
		return 0;
	}

	if (_mode == pitch_bend) {
		return SHRT_MAX;
	}
	
	return 127;
}


int
Sequence::minimum_step_value ()
{
	if (is_raw_midi()) {
		return 0;
	}

	if (_mode == pitch_bend) {
		return SHRT_MIN;
	}
	
	return 0;
}

std::ostream& 
operator<< (std::ostream &ostr, Sequence &seq)
{
	seq.dump (ostr);
	return ostr;
}

std::istream& 
operator>> (std::istream &istr, Sequence &seq)
{
	seq.load (istr);
	return istr;
}

void
Sequence::paste_from (Sequence &seq)
{
	/* This is a little like an assignment operator. However,
	   semantically it is really very different and it would be
	   too risky to allow the compiler to think we are providing
	   operator= when this is nothing like a true operator= would
	   be.

	   Basically, this does nothing to the mode of a sequence,
	   but attempts to do a sane copy of the other data.
	*/

	vector<Step *>::iterator ours;
	vector<Step *>::iterator theirs;

	if ((_mode != pitch_bend && seq.mode() == pitch_bend) ||
	    (_mode == pitch_bend && seq.mode() != pitch_bend)) {
		error << "Cannot paste between pitch-bend and non-pitch-bend "
			 "sequences"
		      << endmsg;
		return;
	}

	if ((is_raw_midi() && !seq.is_raw_midi ()) ||
	    (!is_raw_midi() && seq.is_raw_midi ())) {
		error << "Cannot paste between raw MIDI and non-raw MIDI "
			 "sequences"
		      << endmsg;
		return;
	}

	{

		Glib::Mutex::Lock lm (_lock);

		/* For now, we don't copy very much of the data from the other
		   sequence. This might change.
		*/
		
		if (is_raw_midi()) {
			for (ours = steps.begin(), theirs = seq.steps.begin();
			     ours != steps.end() && theirs != seq.steps.end();
			     ours++, theirs++) {
				
				byte *msg;
				size_t len;
				
				msg = (*ours)->midimsg (&len);
				(*ours)->set (msg, len);
				
				if ((*theirs)->is_enabled ()) {
					(*ours)->enable ();
				} else {
					(*ours)->disable ();
				}
			}		
		} else {
			for (ours = steps.begin(), theirs = seq.steps.begin();
			     ours != steps.end() && theirs != seq.steps.end();
			     ours++, theirs++) {
				
				(*ours)->set ((*theirs)->val());
				
				if ((*theirs)->is_enabled ()) {
					(*ours)->enable ();
				} else {
					(*ours)->disable ();
				}
				
			}
		}
		
		if (makes_sound() && seq.makes_sound ()) {
			set_volume (seq.volume ());
			set_mute (seq.muted ());
		}
		
		/* parameter rows don't get their MIDI ports reset */
		
		if (!is_parameter() && !seq.is_parameter()) {
			_set_read_port (seq.read_port ());
			_set_write_port (seq.write_port ());
			_set_input_channel (seq.input_channel ());
			_set_output_channel (seq.output_channel ());
		}
		
		if (is_controller() && seq.is_controller()) {
			set_controller_id (seq.controller_id());
		}
		
		set_gate_timing (seq.fixed_gate_interval ());
		set_note_length (seq.fixed_note_duration ());
		_set_direction (seq.direction ());
	}

	changed ();
}

void
Sequence::start_midi_grab ()
{
	if (midi_grab_in_progress) {
		error << "MIDI grab already in progress" << endmsg;
		return;
	}

	set_current_step (_first_step);

	_lock.lock ();
	midi_grab_cnt = 0;
	midi_grab_in_progress = true;

	any_connection = input_port->input()->any.connect 
		(mem_fun (*this, &Sequence::handle_midimsg));

	recording_started ();
}

void
Sequence::handle_midimsg (MIDI::Parser &p, byte *buf, size_t bufsize)
{
	byte *msg;
    
	msg = new byte [bufsize];
	memcpy (msg, buf, bufsize);
	steps[midi_grab_cnt]->set (msg, bufsize);
	steps[midi_grab_cnt]->enable ();
	midi_grab_cnt++;

	if (midi_grab_cnt == active_steps()) {
		recording_finished ();
		any_connection.disconnect ();
		midi_grab_in_progress = false;
		midi_grab_cnt = 0;
		_lock.unlock ();
	}
}    

/* Triggers */

void
Sequence::use_trigger (Trigger &trig)
{
	trigger_name = trig.name ();
	trigger_connection = trig.fired.connect (mem_fun (*this, &Sequence::trigger_fired));
	have_trigger = true;
	waiting_for_trigger = true;
	_trigger_mode = single;
	fired_cnt = 0;
}

void
Sequence::drop_trigger ()
{
	trigger_name = "";
	trigger_connection.disconnect ();
	have_trigger = false;
	waiting_for_trigger = false;
	fired_cnt = 0;
}

void
Sequence::set_trigger_mode (TriggerMode mode)
{
	_trigger_mode = mode;
	
	if (mode == off_trigger) {
		waiting_for_trigger = false;
	}
}

void
Sequence::trigger_fired (Trigger &trigger)
{
	Glib::Mutex::Lock lm (_lock);

	if (_trigger_mode != off_trigger) {
		fired_cnt++;
	}

	waiting_for_trigger = false;

	if (_trigger_mode == multretrig) {
		set_current_step (_first_step);
	} else if (_trigger_mode == off_trigger) {
		set_on (false);
	}
}

void
Sequence::set_on (bool yn)
{
	_off = !yn;
	on_off_changed ();
}

void
Sequence::randomize ()
{
	vector<Step *>::iterator i;
	unsigned long upper_value;
	unsigned long lower_value;
	time_t now;

	if (is_raw_midi()) {
		return;
	}

	time (&now);
	srandom ((unsigned int) now);
	upper_value = (long) maximum_step_value ();
	lower_value = (long) minimum_step_value ();

	for (i = steps.begin(); i != steps.end(); i++) {
		unsigned int randval;
		do { 
			randval = random() % upper_value;
		} while (randval < lower_value);

		(*i)->set (randval);
	}
}

void
Sequence::set_steps_on (bool yn)
{
	vector<Step *>::iterator i;

	for (i = steps.begin(); i != steps.end(); i++) {
		if (yn) {
			(*i)->enable ();
		} else {
			(*i)->disable ();
		}
	}
}

void
Sequence::mode_fill (MusicalMode::Type type, int octaves)
{	
	vector<Step *>::iterator i;

	if (type == MusicalMode::Random) {
		randomize ();
		return;
	}

	if (!makes_sound() || steps.size() < 2) {
		return;
	}
	
	MusicalMode m (type);
	size_t lim;
	int base;
	vector<float>::iterator interval;

	lim = m.steps.size();
	
	i = steps.begin();
	base = (*i)->val();
	++i;

	interval = m.steps.begin();

	for (; i != steps.end(); i++) {

		if (interval == m.steps.end()) {
			
			/* last distance was the end of the scale,
			   so wrap, adding the next note at one
			   octave above the last root.
			*/
			
			interval = m.steps.begin();
			base = base + 12;
			(*i)->set (base);
		} else {
			(*i)->set ((size_t) floor (base + (2.0 * (*interval))));
			interval++;
		}
	}
}
