/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: sequence.h,v 1.5 2000/05/03 15:54:21 pbd Exp $
*/

#ifndef __sw_sequence_h__
#define __sw_sequence_h__


class Sequencer;
class Step;
class ostream;
class istream;

#include <glibmm/thread.h>
#include <sigc++/signal.h>
#include <midi++/port.h>
#include <midi++/channel.h>

#include "mode.h"
#include "types.h"
#include "trigger.h"

class Sequence : public sigc::trackable {
  public:

	enum ModeBits {
		sound = 0x1,    /* noise is guaranteed to be made */
		channel = 0x2,  /* channel message sent */
		midi = 0x4,     /* raw MIDI data */
		parameter = 0x8, /* no message sent - controls other row */
		chord = 0x10,    /* more than one note "simultaneously" */
		relative = 0x20,  /* transposes on MIDI NoteOn */
		note = 0x40,      /* generates noteOn or noteOff */
		control_msg = 0x80  /* generates a controller message */
	};

	enum Mode {
		absolute_pitch      = 0, 
		relative_pitch      = 1,
		single_pitch        = 2,
		midibytes           = 3,
		recorded_midibytes  = 4, 
		gate_timing         = 5,
		note_length         = 6,
		note_on_velocity    = 7,
		note_off_velocity   = 8,
		absolute_chord      = 9,
		relative_chord      = 10,
		program_change      = 11,
		controller          = 12,
		pitch_bend          = 13,
		polytouch           = 14,
		aftertouch          = 15,
	};

	static const char *mode_strings[];

	enum Direction {
		forwards = 0, 
		backwards = 1, 
		end_to_end = 2,
		rd_random = 3
	};

	static const char *direction_strings[];

	enum TriggerMode { 
		single = 0,
		multiple = 1,
		multretrig = 2,
		off_trigger = 3
	};

	static const char *trigger_mode_strings[];

	Sequence (size_t id, size_t numsteps, Sequencer &myseq);
	~Sequence ();

	int id() { return _id; }
	Sequencer &sequencer() { return _sequencer; }

	Step *step (size_t n)   { return steps[n]; }
	size_t last_step ()     { return _last_step; }
	size_t first_step ()    { return _first_step; }
	size_t start_step ()    { return _start_step; }
	size_t current_step ()  { return _current_step; }

	void use_trigger (Trigger &t);
	void drop_trigger ();
	TriggerMode trigger_mode() const { return _trigger_mode; }
	void set_trigger_mode (TriggerMode);
	std::string get_trigger_name() const { return trigger_name; }

	void set_default_step_values (bool preserve);
	void set_base_note (int note, bool absolute_note = 1);
	void set_all_steps (int note);
	void set_current_step (size_t step);
	void set_first_step (size_t step);
	void set_last_step (size_t step);
	void set_start_step (size_t step);

	void set_read_port (MIDI::Port *p);
	void set_write_port (MIDI::Port *p);

	/* Control-oriented stuff */

	int makes_sound () {
		return _modebits & (unsigned char)Sequence::sound; }
	int is_raw_midi () {
		return _modebits & (unsigned char)Sequence::midi; }
	int is_relative () {
		return _modebits & (unsigned char)Sequence::relative; }
	int is_absolute () {
		return !is_relative(); }
	int is_chord    () {
		return _modebits & (unsigned char)Sequence::chord; }
	int is_note     () {
		return _modebits & (unsigned char)Sequence::note; }
	int is_channel_msg () {
		return _modebits & (unsigned char)Sequence::channel; }
	int is_parameter () {
		return _modebits & (unsigned char)Sequence::parameter; }
	int is_controller () {
		return _modebits & (unsigned char)Sequence::control_msg; }

	void add_controlled_row (int row) {
		controlled_rows |= (1 << row);
	} 

	void remove_controlled_row (int row) {
		controlled_rows &= ~(1 << row);
	}

	int controls_row (int r) {
		return controlled_rows & (1<< r);
	}

	MIDI::byte      volume ()        { return _volume; }
	bool            silent ()        { return _volume == 0 || _muted ; }
	Direction       direction ()     { return _direction; }
	MIDI::channel_t input_channel () { return _input_channel; }
	MIDI::channel_t output_channel (){ return _output_channel; }
	MIDI::byte      controller_id()  { return _controller_id; }
	Sequence::Mode  mode()           { return _mode; }
	MIDI::Port     *read_port()      { return input_port; }
	MIDI::Port     *write_port()     { return output_port; }
	bool            skipping()       { return _skip; }
	bool            muted()          { return _muted; }
	bool            off ()           { return _off; }
	bool            on ()            { return !_off; }

	void           set_skip (bool s) { _skip = s; }
	void           set_mode (int);
	void           set_volume (MIDI::byte v) { _volume = v; }
	void           set_mute (bool m) { _muted = m; }
	void           set_input_channel (MIDI::channel_t ch);
	void           set_output_channel (MIDI::channel_t ch);
	void           set_gate_timing (ticks_t t) { gate_time = t; }
	void           set_note_length (ticks_t t) { note_time = t; }
	void           set_next_gate   (ticks_t t);
	void           set_controller_id (MIDI::byte cid) { _controller_id = cid; }
	void           set_direction (Direction d);
	void           set_on (bool);
	void           set_steps_on (bool);

	void           randomize ();
	void           mode_fill (MusicalMode::Type, int octaves);

	bool acquire_lock () { return _lock.trylock(); }
	void release_lock () { _lock.unlock(); }

	int maximum_step_value ();
	int minimum_step_value ();

	sigc::signal0<void> on_off_changed;
	sigc::signal0<void> changed;
	sigc::signal0<void> mode_changed;
	sigc::signal0<void> input_midi_changed;
	sigc::signal0<void> output_midi_changed;
	sigc::signal0<void> recording_started;
	sigc::signal0<void> recording_finished;

	/* This is a bit like an assignment operator, but doesn't have
	   full assignment semantics.
	*/

	void paste_from (Sequence &);

  protected:
	void dump (std::ostream &);
	void load (std::istream &);
	friend std::ostream& operator<< (std::ostream &, Sequence &);
	friend std::istream& operator>> (std::istream &, Sequence &);

  protected:
	friend class Sequencer;
	void get_next_wakeup (ticks_t& current_wakeup);
	void execute (size_t current_tick);
	void check (unsigned int when);
	void advance ();
	void reset ();
	bool needs_attention (ticks_t when) {
		return !_off && next_gate <= when;
	}

  protected:
	friend class SequenceUI;
	Step *new_step ();
	ticks_t  fixed_gate_interval () { return gate_time; }
	ticks_t  fixed_note_duration () { return note_time; }

  protected:
	friend class Step;

	/* methods for emitting MIDI messages */

	void note_on  (MIDI::byte note);
	void note_off (MIDI::byte note);

	int midimsg (MIDI::byte *msg, unsigned int len) {
		return output_port->midimsg (msg, len);
	} 

	int channel_msg (MIDI::channel_t c, MIDI::byte id, 
			 MIDI::byte val1, MIDI::byte val2) {
		
		return output_port->channel(c)->channel_msg (id, val1, val2);
	}
	
	int all_notes_off (MIDI::channel_t channel) {
		return output_port->channel (channel)->all_notes_off ();
	}
	
	int control (MIDI::channel_t channel, MIDI::byte id, MIDI::byte value) {
		return output_port->channel (channel)->control (id, value);
	}

  private:
	enum E2EMode { 
		e2e_forward,
		e2e_backward,
	};

	std::vector<Step *>	steps;
	int			_id;
	size_t			_last_step; 
	size_t			_first_step;
	size_t			_current_step;
	size_t                  _start_step;
	bool                    _off;      /* not running */
	bool			_skip;     /* skip disabled steps ? */
	short			base_note; /* root note for transposition */
	ticks_t			next_gate;
	Sequencer		&_sequencer;
	MIDI::Port		*input_port;
	MIDI::Port		*output_port;
	MIDI::channel_t		_input_channel;
	MIDI::channel_t		_output_channel;
	MIDI::byte		_controller_id;
	unsigned int		gate_time;
	unsigned int		note_time;
	unsigned int		controlled_rows; /* mask for up to 32 rows */
	MIDI::byte		_volume;
	bool			_muted;
	Mode			_mode;
	unsigned char		_modebits;
	E2EMode			e2emode;
	Direction		_direction;
	sigc::connection	note_on_connection;

	/* How we monitor Recorded MIDI mode */

	sigc::connection        any_connection;
	bool                    midi_grab_in_progress;
	size_t                  midi_grab_cnt;
	void                    start_midi_grab ();
	void                    handle_midimsg (MIDI::Parser &, 
						MIDI::byte *buf, 
						size_t bufsize);
	

	Glib::Mutex _lock;

	/* Handling triggers */

	void trigger_fired (Trigger &);
	sigc::connection trigger_connection;
	unsigned int fired_cnt;
	bool have_trigger;
	bool waiting_for_trigger;
	TriggerMode _trigger_mode;
	void reset_trigger ();
	std::string trigger_name;

	void on_notes_off ();
	void play_chord (int onoff, ticks_t when);

	/* computing important stuff */

	MIDI::byte on_velocity ();
	MIDI::byte off_velocity ();
	ticks_t    gate_interval ();
	ticks_t    note_duration ();

	/* handling MIDI-relative modes */

	void maybe_listen_to_note_on ();
	void handle_note_on (MIDI::Parser &, MIDI::EventTwoBytes *);

	/* miscellany */

	void foreach_step (void (*)(Step &, void *), void *);
	size_t active_steps()   { return _last_step - _first_step + 1; }


	/* lock free variants */

	void _set_mode (int);
	void _set_input_channel (MIDI::channel_t ch);
	void _set_output_channel (MIDI::channel_t ch);
	void _set_next_gate   (ticks_t t);
	void _set_direction (Direction d);
	void _set_on (bool);
	void _set_steps_on (bool);
	void _set_start_step (size_t step);
	void _set_base_note (int note, bool absolute_note = 1);
	void _set_read_port (MIDI::Port *p);
	void _set_write_port (MIDI::Port *p);
};

std::ostream& operator<< (std::ostream &, Sequence &);
std::istream& operator>> (std::istream &, Sequence &);


#endif /* __sw_sequence_h__ */
