#include <limits.h>

#include <pbd/error.h>
#include <midi++/port.h>
#include <midi++/manager.h>
#include <midi++/types.h>
#include <gtkmm2ext/utils.h>

#include "ui.h"
#include "sequence_ui.h"
#include "sequencer.h"
#include "sequence.h"
#include "step_ui.h"

#include "i18n.h"

using namespace PBD;

vector<string> SequenceUI::mode_strings;
vector<string> SequenceUI::direction_strings;

bool SequenceUI::have_descriptive_strings = false;

static const char *longest_mode_string = "";
static const char *longest_direction_string = "";

using namespace MIDI;
using namespace Gtk;

void
SequenceUI::get_descriptive_strings ()

{
	have_descriptive_strings = true;
	int n;

	for (n = 0; Sequence::mode_strings[n]; n++) {
		const char *str;

		str = Sequence::mode_strings[n];
		
		if (strlen (str) > strlen (longest_mode_string)) {
			longest_mode_string = str;
		}
		mode_strings.push_back ((char *) str);
	}

	for (n = 0; Sequence::direction_strings[n]; n++) {
		const char *str;

		str = Sequence::direction_strings[n];
		
		if (strlen (str) > strlen (longest_direction_string)) {
			longest_direction_string = str;
		}
		direction_strings.push_back ((char *) str);
	}
}

SequenceUI::SequenceUI (Sequence &seq) 
	: _sequence (seq),
	  volumeAdjustment (seq.volume(), 0, 127, 1, 5, 0),
	  volumeControl (SequencerUI::instance()->knob_pixbuf, _sequence, &volumeAdjustment, false),
	  startAdjustment (0, 0, 999, 1, 10),
	  startSpin (&startAdjustment, "StartSelector"),
	  gateAdjustment (5, 1, 1000, 1, 10),
	  gateSpin (&gateAdjustment, "GateSelector"),
	  noteAdjustment (5, 1, 1000, 1, 10),
	  noteSpin (&noteAdjustment, "NoteSelector"),
	  inchanAdjustment (1, 1, 16, 1, 4),
	  outchanAdjustment (1, 1, 16, 1, 4),
	  inchanSelector (inchanAdjustment),
	  outchanSelector (outchanAdjustment)
	   
{
	const char *longest_string;

	if (!have_descriptive_strings) {
		get_descriptive_strings ();
	}

	ignore_state_changes = 0;

	_sequence.mode_changed.connect (mem_fun (*this, &SequenceUI::map_mode));
	_sequence.on_off_changed.connect (mem_fun (*this, &SequenceUI::map_on_off));
	_sequence.changed.connect (mem_fun (*this, &SequenceUI::map_state));
	_sequence.recording_started.connect (mem_fun (*this, &SequenceUI::recording_start));
	_sequence.recording_finished.connect (mem_fun (*this, &SequenceUI::recording_end));

	sequence_label.set_name ("SequenceLabel");
	hsep1.set_name ("SequenceSeparator");
	hsep2.set_name ("SequenceSeparator");

	labelSeparatorBox.set_spacing (5);
	labelSeparatorBox.pack_start (hsep1);
	labelSeparatorBox.pack_start (sequence_label, false, false);
	labelSeparatorBox.pack_start (hsep2);

	pack_start (labelSeparatorBox, false, false);
	pack_start (hpacker, false, false);

	hpacker.set_homogeneous (false);
	hpacker.set_spacing (2);

	volumeBox.set_homogeneous (false);
	volumeBox.set_spacing (2);

	muteLabel.set_text ("mute");
	muteLabel.set_name ("ModeButton");
	muteButton.set_name ("ModeButton");
	muteButton.add (muteLabel);

	SequencerUI::instance()->set_tip (&volumeControl, _("(Base) volume of notes in this sequence"));

	volumeBox.pack_start (volumeControl, false, false);
	volumeBox.pack_end (muteButton, false, false);

	muteButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::mute));

	hpacker.pack_start (stepScroller, true, true, 10);
	hpacker.pack_start (volumeBox, false, false);

	step_hpacker.set_spacing (1);
	step_hpacker.set_homogeneous (false);
	stepScroller.set_name ("StepScroller");
	stepScroller.set_policy (Gtk::POLICY_AUTOMATIC, Gtk::POLICY_NEVER);
	stepScroller.add (step_hpacker);

	if (strlen (longest_direction_string) > 
	    strlen (longest_mode_string)) {
		longest_string = longest_direction_string;
	} else {
		longest_string = longest_mode_string;
	}
	
	/* The Mode Selector */

	Gtkmm2ext::set_popdown_strings (modeSelector, mode_strings);
	modeSelector.set_name ("ModeSelector");
	modeSelector.signal_changed().connect (mem_fun (*this, &SequenceUI::set_sequence_mode));
	SequencerUI::instance()->set_tip (&modeSelector, _("Type of data generated by this sequence"));

	/* The Direction selector */

	Gtkmm2ext::set_popdown_strings (directionSelector, direction_strings);
	directionSelector.set_name ("DirectionSelector");
	directionSelector.signal_changed().connect (mem_fun (*this, &SequenceUI::set_sequence_direction));
	SequencerUI::instance()->set_tip (&directionSelector, _("Direction this sequence executes in"));

	/* package up mode, direction */

	modeControlBox.set_spacing (2);
	modeControlBox.pack_start (modeSelector, true, true);
	modeControlBox.pack_start (directionSelector, true, true);

	/* The Gate Interval control widget */

	gateSpin.set_print_func (integer_filter, 0);
	Gtkmm2ext::set_size_request_to_display_given_text (gateSpin, "1000", 15, 5);
	gateFrame.add (gateSpin);
	gateFrame.set_shadow_type (Gtk::SHADOW_IN);
	gateFrame.set_name ("GateSelector");
	gateBox.set_homogeneous (false);
	gateBox.set_spacing (2);
	gateBox.pack_start (gateFrame, false, false);
	gateBoxLabel.set_text ("Gate");
	gateBoxLabel.set_name ("TimingBoxLabel");
	gateBoxLabel.set_alignment (1.0, 0.5);
	gateBox.pack_start (gateBoxLabel, false, false);
	gateSpin.get_adjustment().signal_value_changed().connect(mem_fun (*this,&SequenceUI::set_gate_interval));
	SequencerUI::instance()->set_tip (&gateSpin, _("Interval between steps for this sequence, in ticks"));

	/* The Note Duration control widget */

	noteSpin.set_print_func (integer_filter, 0);
	Gtkmm2ext::set_size_request_to_display_given_text (noteSpin, "1000", 15, 5);
	noteFrame.add (noteSpin);
	noteFrame.set_shadow_type (Gtk::SHADOW_IN);
	noteFrame.set_name ("NoteSelector");
	noteBox.set_homogeneous (false);
	noteBox.set_spacing (2);
	noteBox.pack_start (noteFrame, false, false);
	noteBoxLabel.set_text ("Note");
	noteBoxLabel.set_name ("TimingBoxLabel");
	noteBoxLabel.set_alignment (1.0, 0.5);
	noteBox.pack_start (noteBoxLabel, false, false);
	noteSpin.get_adjustment().signal_value_changed().connect(mem_fun (*this,&SequenceUI::set_note_duration));
	SequencerUI::instance()->set_tip (&noteSpin, _("Duration of notes in this sequence, in ticks"));

	/* The start-step-offset adjuster */

	startSpin.set_print_func (integer_filter, 0);
	Gtkmm2ext::set_size_request_to_display_given_text (startSpin, "1000", 15, 5);
	startFrame.add (startSpin);
	startFrame.set_shadow_type (Gtk::SHADOW_IN);
	startFrame.set_name ("StartSelector");
	startBox.set_homogeneous (false);
	startBox.set_spacing (2);
	startBoxLabel.set_text ("Skew");
	startBoxLabel.set_name ("TimingBoxLabel");
	startBoxLabel.set_alignment (1.0, 0.5);
	startBox.pack_start (startFrame, false, false);
	startBox.pack_start (startBoxLabel, false, false);
	startSpin.get_adjustment().signal_value_changed().connect(mem_fun (*this,&SequenceUI::set_start_step));
	SequencerUI::instance()->set_tip (&startSpin, _("Number of ticks to delay start of this sequence"));
	
	/* pack up the gate, note widgets with labels */

	gate_and_noteBox.set_spacing (2);
	gate_and_noteBox.pack_start (gateBox, false, false);
	gate_and_noteBox.pack_start (noteBox, false, false);
	gate_and_noteBox.pack_start (startBox, false, false);

	/* The MIDI Channel control widgets */

	outchanSelector.get_adjustment()->signal_value_changed().connect (mem_fun (*this,&SequenceUI::set_output_channel));
	SequencerUI::instance()->set_tip (&outchanSelector, _("Output channel"));
	outchanSelector.set_digits (0);
	outchanSelector.set_name ("ChannelSelector");

	inchanSelector.get_adjustment()->signal_value_changed().connect (mem_fun (*this,&SequenceUI::set_input_channel));
	SequencerUI::instance()->set_tip (&inchanSelector, _("Input channel"));
	inchanSelector.set_digits (0);
	inchanSelector.set_name ("ChannelSelector");

	/* The MIDI input port selector */

	Gtkmm2ext::set_popdown_strings (inportSelector, SequencerUI::instance()->port_strings());
	inportBoxLabel.set_text ("IN");
	inportBoxLabel.set_alignment (1.0, 0.5);
	inportBoxLabel.set_name ("MIDILabel");
	inportBox.set_homogeneous (false);
	inportBox.set_spacing (2);
	inportBox.pack_start (inportSelector, false, false);
	inportBox.pack_start (inchanSelector, false, false);
	inportBox.pack_start (inportBoxLabel, false, false, 2);
	inportSelector.signal_changed().connect (mem_fun (*this,&SequenceUI::set_input_port));
	SequencerUI::instance()->set_tip (&inportSelector, _("Input port"));

	/* The MIDI output port selector */

	Gtkmm2ext::set_popdown_strings (outportSelector, SequencerUI::instance()->port_strings());
	outportBoxLabel.set_text ("OUT");
	outportBoxLabel.set_alignment (1.0, 0.5);
	outportBoxLabel.set_name ("MIDILabel");
	outportBox.set_homogeneous (false);
	outportBox.set_spacing (2);
	outportBox.pack_start (outportSelector, false, false);
	outportBox.pack_start (outchanSelector, false, false);
	outportBox.pack_start (outportBoxLabel, false, false, 2);
	outportSelector.signal_changed().connect (mem_fun (*this,&SequenceUI::set_output_port));
	SequencerUI::instance()->set_tip (&outportSelector, _("Output port"));

	/* pack up port and channel selectors */

	midi_selectionBox.set_homogeneous (false);
	midi_selectionBox.set_spacing (2);
	midi_selectionBox.pack_start (inportBox, false, false);
	midi_selectionBox.pack_start (outportBox, false, false);

	/* pack the above into the top part of the right hand side */

	rightUpperBox.set_homogeneous (false);
	rightUpperBox.set_spacing (2);
	rightUpperBox.pack_start (modeControlBox, false, false);
	rightUpperBox.pack_start (gate_and_noteBox, false, false);
	rightUpperBox.pack_start (midi_selectionBox, false, false);

	/* now pack the mode buttons, track control buttons, 
	   controller selector and trigger button
	   into the lower part of the right hand side 
	*/

	/* Mode control buttons */

	runLabel.set_text ("run");
	runLabel.set_name ("ModeButton");
	runButton.set_name ("ModeButton");
	runButton.add (runLabel);

	addStepLabel.set_text ("add step");
	addStepLabel.set_name ("ModeButton");
	addStepButton.set_name ("ModeButton");
	addStepButton.add (addStepLabel);

	offLabel.set_text ("off");
	offLabel.set_name ("ModeButton");
	offButton.set_name ("ModeButton");
	offButton.add (offLabel);

	skipLabel.set_text ("skip");
	skipLabel.set_name ("ModeButton");
	skipButton.set_name ("ModeButton");
	skipButton.add (skipLabel);

	copyLabel.set_text ("copy");
	copyLabel.set_name ("ModeButton");
	copyButton.set_name ("ModeButton");
	copyButton.add (copyLabel);

	pasteLabel.set_text ("paste");
	pasteLabel.set_name ("ModeButton");
	pasteButton.set_name ("ModeButton");
	pasteButton.add (pasteLabel);

	modefillLabel.set_text ("mode fill");
	modefillLabel.set_name ("ModeButton");
	modefillButton.set_name ("ModeButton");
	modefillButton.add (modefillLabel);
	SequencerUI::instance()->set_tip (&modefillButton, _("Fill sequence with selected mode"));

	triggerLabel.set_text ("trigger");
	triggerLabel.set_name ("ModeButton");
	triggerButton.set_name ("ModeButton");
	triggerButton.add (triggerLabel);
	SequencerUI::instance()->set_tip (&triggerButton, _("Manage triggering for this sequence"));
	
	trigger_list_menu = 0;
	trigger_menu = 0;
	modefill_menu = 0;
	
	mode_buttonBox.set_homogeneous (false);
	mode_buttonBox.set_spacing (2);
	mode_buttonBox.pack_start (runButton, false, false);
	mode_buttonBox.pack_start (addStepButton, false, false);
	mode_buttonBox.pack_start (skipButton, false, false);
	mode_buttonBox.pack_start (offButton, false, false);
	mode_buttonBox.pack_start (copyButton, false, false);
	mode_buttonBox.pack_start (pasteButton, false, false);
	mode_buttonBox.pack_start (modefillButton, false, false);
	mode_buttonBox.pack_start (triggerButton, false, false);

	runButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::run));
	skipButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::skip));
	copyButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::copy));
	pasteButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::paste));
	addStepButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::new_step));
	offButton.signal_toggled().connect(mem_fun (*this,&SequenceUI::off));
	triggerButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::trigger_click));
	modefillButton.signal_clicked().connect(mem_fun (*this,&SequenceUI::modefill_click));

	/* The MIDI Controller ID selector (often hidden) */

	Gtkmm2ext::set_popdown_strings (controllerSelector, SequencerUI::controller_strings());
	controllerSelector.set_name ("ControllerSelector");
	Gtkmm2ext::set_size_request_to_display_given_text (controllerSelector, SequencerUI::longest_controller_string, 15, 0);
	controllerSelector.signal_changed().connect (mem_fun (*this, &SequenceUI::set_controller_id));
	controllerBox.pack_start (controllerSelector, false, false);
	
	row_and_triggerBox.set_homogeneous (false);
	row_and_triggerBox.set_spacing (10);
	row_and_triggerBox.pack_start (controllerBox, false,false);

	/* buttons for the track box come later */
	
	row_and_triggerBox.pack_start (trackBox, false, false);

	rightUpperBox.set_homogeneous (false);
	rightUpperBox.set_spacing (10);
	rightLowerBox.pack_start (mode_buttonBox, false, false);
	rightLowerBox.pack_start (row_and_triggerBox, false, false);

	/* OK, stuff the two parts of the RHS together */

	rightBox.set_spacing (2);
	rightBox.pack_start (rightUpperBox, false, false);
	rightBox.pack_end (rightLowerBox, false, false);

	/* and stuff it on the RHS */

	hpacker.pack_start (rightBox, false, true);

	/* now add step UI's */

	_sequence.foreach_step (SequenceUI::_add_step, this);

	show_all ();

	/* manage notifications that our sequence is not the copy sequence
	*/

	Sequencer::instance()->copy_sequence_changed.connect 
		(mem_fun (*this, &SequenceUI::copy_status_change));

	/* cleanup that last step to reflect the current mode */

	map_state ();
}

void
SequenceUI::map_state ()

{
	vector<StepUI *>::iterator i;
	vector<Gtk::ToggleButton *>::iterator tb;
	char buf[32];
	int n;

	ignore_state_changes++;

	/* first take care of sequence level things */

	sprintf (buf, "Sequence %d", _sequence.id() + 1);
	sequence_label.set_text (buf);

	map_mode ();
	map_on_off ();

	muteButton.set_active (_sequence.muted ());
	skipButton.set_active (_sequence.skipping ());
	copyButton.set_active (_sequence.sequencer().copy_sequence() == _sequence.id());
			       
	volumeControl.set_value (_sequence.volume ());
	gateSpin.get_adjustment().set_value (_sequence.fixed_gate_interval());
	noteSpin.get_adjustment().set_value (_sequence.fixed_note_duration());

	if (_sequence.is_controller()) {
		controllerSelector.set_active_text (MIDI::controller_names[_sequence.controller_id()]);
	}

	outportSelector.set_active_text (_sequence.write_port()->name());
	outchanSelector.get_adjustment()->set_value (_sequence.output_channel() + 1);

	inportSelector.set_active_text (_sequence.read_port()->name());
	inchanSelector.get_adjustment()->set_value (_sequence.input_channel() + 1);

	directionSelector.set_active_text (Sequence::direction_strings[(int) _sequence.direction()]);
	modeSelector.set_active_text (Sequence::mode_strings[(int) _sequence.mode()]);

	for (n = 0, tb = trackButtons.begin ();
	     tb != trackButtons.end ();
	     tb++, n++) {
		if (_sequence.controls_row(n)) {
			(*tb)->set_active (true);
		} else {
			(*tb)->set_active (false);
		}
	}

	ignore_state_changes--;

	/* map each step */

	for (i = steps.begin (); i != steps.end(); i++) {
		(*i)->map_state ();
	}
}

void
SequenceUI::map_on_off ()

{
	ignore_state_changes++;

	if (_sequence.on()) {
		runButton.set_active (true);
	} else {
		runButton.set_active (false);
	}

	ignore_state_changes--;
}

void
SequenceUI::map_mode ()

{
	vector<StepUI *>::iterator step;
	Sequence::Mode mode;

	mode = _sequence.mode ();

	/* only one mode uses this, so get rid of it as the default */

	controllerBox.hide ();

	/* almost everything cares about this, so show it as the default */

	outportBox.show ();

	switch (mode) {
	case Sequence::recorded_midibytes:
	case Sequence::midibytes:
		outchanSelector.hide ();
		trackBox.hide ();
		triggerButton.show ();
		break;

	case Sequence::controller:
		controllerBox.show_all ();
		/* FALLTHRU */
	case Sequence::pitch_bend:
		trackBox.hide ();
		triggerButton.show ();
		break;

	case Sequence::note_on_velocity:
	case Sequence::note_off_velocity:
	case Sequence::gate_timing:
	case Sequence::note_length:
		outportBox.hide ();
		trackBox.show_all ();
		triggerButton.hide ();
		break;

	case Sequence::absolute_chord:
	case Sequence::relative_chord:
		trackBox.hide ();
		triggerButton.show_all ();
		break;

	default:
		midi_selectionBox.show_all();
		controllerBox.hide ();
		triggerButton.show_all ();
		trackBox.hide ();
	}

	if (_sequence.makes_sound()) {
		volumeBox.show_all ();
		noteBox.show ();
	} else {
		volumeBox.hide ();
		noteBox.hide ();
	}

	for (step = steps.begin(); step != steps.end(); step++) {
		(*step)->map_state ();
	}
}

void
SequenceUI::_add_step (Step &step, void *arg)

{
	SequenceUI *sui = (SequenceUI *) arg;
	
	sui->add_step (step);
}

void
SequenceUI::add_step (Step &step)

{
	StepUI *stepui;

	stepui = new StepUI (step);

	step_hpacker.pack_start (*stepui, false, false);
	stepui->show_all ();
	steps.push_back (stepui);
}

void
SequenceUI::touch ()

{
	vector<StepUI *>::iterator s;

	// std::cerr << "touch sequence " << _sequence.id() << "\n";
	
	for (s = steps.begin(); s != steps.end(); s++) {
		(*s)->touch ();
	}
}

void
SequenceUI::set_sequence_mode ()
{
	string new_mode;
	Sequence::Mode mode = Sequence::absolute_chord; /* keep gcc happy */

	new_mode = modeSelector.get_active_text ();

	if (new_mode == "Absolute Pitch") {
		mode = Sequence::absolute_pitch;
	} else if (new_mode == "Relative Pitch") {
		mode = Sequence::relative_pitch;
	} else if (new_mode == "Single Pitch (drum)") {
		mode = Sequence::single_pitch;
	} else if (new_mode == "MIDI bytes") {
		mode = Sequence::midibytes;
	} else if (new_mode == "Recorded MIDI") {
		mode = Sequence::recorded_midibytes;
	} else if (new_mode == "Gate Timing") {
		mode = Sequence::gate_timing;
	} else if (new_mode == "Note Length") {
		mode = Sequence::note_length;
	} else if (new_mode == "On Velocity") {
		mode = Sequence::note_on_velocity;
	} else if (new_mode == "Off Velocity") {
		mode = Sequence::note_off_velocity;
	} else if (new_mode == "Absolute Chord") {
		mode = Sequence::absolute_chord;
	} else if (new_mode == "Relative Chord") {
		mode = Sequence::relative_chord;
	} else if (new_mode == "Program Change") {
		mode = Sequence::program_change;
	} else if (new_mode == "Controller") {
		mode = Sequence::controller;
	} else if (new_mode == "Pitch Bend") {
		mode = Sequence::pitch_bend;
	} else if (new_mode == "Polytouch") {
		mode = Sequence::polytouch;
	} else if (new_mode == "Aftertouch") {
		mode = Sequence::aftertouch;
	} else {
		fatal << "Unknown sequence mode string \""
		      << new_mode
		      << '"'
		      << endmsg;
		/*NOTREACHED*/
	}

	_sequence.set_mode (mode);
}

void
SequenceUI::set_sequence_direction ()
{
	if (ignore_state_changes) {
		return;
	}

	string str = directionSelector.get_active_text ();
	Sequence::Direction new_dir = Sequence::forwards; /* keep gcc happy */

	if (str == "Forwards") {
		new_dir = Sequence::forwards;
	} else if (str == "Backwards") {
		new_dir = Sequence::backwards;
	} else if (str == "End-to-end") {
		new_dir = Sequence::end_to_end;
	} else if (str == "Random") {
		new_dir = Sequence::rd_random;
	} else {
		fatal << "unknown direction string \""
		      << str
		      << '"'
		      << endmsg;
		/*NOTREACHED*/
	}

	_sequence.set_direction (new_dir);
}

void
SequenceUI::mute ()

{
	if (ignore_state_changes) {
		return;
	}
	_sequence.set_mute (!_sequence.muted());

}

void
SequenceUI::run ()

{
	bool seqon;

	if (ignore_state_changes) {
		return;
	}

	seqon = _sequence.on ();

	if (runButton.get_active()) {
		if (!seqon) {
			_sequence.set_on (true);
		}
	} else {
		if (seqon) {
			_sequence.set_on (false);
		}
	}
}

void
SequenceUI::skip ()

{
	if (ignore_state_changes) {
		return;
	}

	_sequence.set_skip (!_sequence.skipping());
}

void
SequenceUI::copy ()

{
	if (ignore_state_changes) {
		return;
	}

	if (copyButton.get_active()) {
		_sequence.sequencer().set_copy_sequence (_sequence.id());
	} else {
		_sequence.sequencer().unset_copy_sequence (_sequence.id());
	}
}

void
SequenceUI::paste ()

{
	if (ignore_state_changes) {
		return;
	}

	_sequence.sequencer().paste_sequence (_sequence.id());
}

void
SequenceUI::set_gate_interval ()

{
	if (ignore_state_changes) {
		return;
	}

	_sequence.set_gate_timing
		((unsigned long) gateSpin.get_adjustment().get_value());
}

void
SequenceUI::set_note_duration ()

{
	if (ignore_state_changes) {
		return;
	}

	_sequence.set_note_length
		((unsigned long) noteSpin.get_adjustment().get_value());
}

void
SequenceUI::set_start_step ()

{
	if (ignore_state_changes) {
		return;
	}

	_sequence.set_start_step
		((unsigned long) startSpin.get_adjustment().get_value());
}

void
SequenceUI::set_controller_id ()
{
	if (ignore_state_changes) {
		return;
	}

	Glib::ustring str = controllerSelector.get_active_text ();
	byte id = MIDI::decode_controller_name (str.c_str());
	_sequence.set_controller_id (id);
}

void
SequenceUI::set_input_port ()
{
	MIDI::Port *port;

	if (ignore_state_changes) {
		return;
	}

	if ((port = MIDI::Manager::instance()->port (inportSelector.get_active_text())) == 0) {
		fatal << "unknown MIDI input port " << inportSelector.get_active_text() << endmsg;
	}
	_sequence.set_read_port (port);
}

void
SequenceUI::set_output_port ()
{
	MIDI::Port *port;

	if (ignore_state_changes) {
		return;
	}

	if ((port = MIDI::Manager::instance()->port (outportSelector.get_active_text())) == 0) {
		fatal << "unknown MIDI output port " << outportSelector.get_active_text() << endmsg;
	}
	_sequence.set_write_port (port);
}

void
SequenceUI::set_input_channel ()
{
	channel_t chn;

	if (ignore_state_changes) {
		return;
	}
	chn = (channel_t) outchanSelector.get_adjustment()->get_value () - 1;

	_sequence.set_input_channel (chn);
}

void
SequenceUI::set_output_channel ()
{
	channel_t chn;

	if (ignore_state_changes) {
		return;
	}

	chn = (channel_t) outchanSelector.get_adjustment()->get_value () - 1;
	_sequence.set_output_channel (chn);
}

void
SequenceUI::new_step ()

{
	Step   *nstep;

	if ((nstep = _sequence.new_step ()) == 0) {
		return;
	}

	add_step (*nstep);
}

void
SequenceUI::add_track_button (unsigned int id)

{
	Gtk::ToggleButton *nb;
	Gtk::VBox *vb;
	char buf[32];
	Gtk::Label *nl;

	/* a little recursion never hurt anyone: make sure we always
	   add all buttons for all rows.  
	*/

	if (id > trackButtons.size()) {
		add_track_button (id - 1);
	}

	sprintf (buf, "%d", id + 1);
	nl = new Gtk::Label (buf);
	nb = new Gtk::ToggleButton;
	nb->add (*nl);

	nl->set_name ("TrackControlButton");
	nb->set_name ("TrackControlButton");

	trackButtons.push_back (nb);

	/* pack into a vbox to limit vertical growth - we want the
	   button to be as small as it can be, given its label (and
	   thus style).
	*/

	vb = new Gtk::VBox;
	vb->pack_start (*nb, false, false);
	trackBox.pack_start (*vb, false, false);

	vb->show_all ();

	nb->signal_toggled().connect(sigc::bind(mem_fun (*this,&SequenceUI::track_toggle),id));
}

void
SequenceUI::track_toggle (unsigned int id)

{
	if (ignore_state_changes) {
		return;
	}

	if (trackButtons[id]->get_active()) {
		_sequence.add_controlled_row (id);
	} else {
		_sequence.remove_controlled_row (id);
	}
}

void
SequenceUI::copy_status_change (int current_copy_sequence)

{
	ignore_state_changes = true;

	if ((int) _sequence.id() == current_copy_sequence) {
		copyButton.set_active (true);
	} else {
		copyButton.set_active (false);
	}

	ignore_state_changes = false;
}

void
SequenceUI::recording_start ()

{
	/* XXX what to do ? */
}

void
SequenceUI::recording_end ()

{
	gdk_beep ();
}

void
SequenceUI::off ()
{
	if (offButton.get_active()) {
		_sequence.set_steps_on (false);
	} else {
		_sequence.set_steps_on (true);
	}
}

void
SequenceUI::build_modefill_menu ()
{
	using namespace Menu_Helpers;

	modefill_menu = manage (new Menu());
	MenuList& items = modefill_menu->items();

	modefill_menu->set_name ("ModeFillMenu");
				     
	items.push_back (MenuElem (_("Random"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Random)));
	items.push_back (MenuElem (_("Dorian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Dorian)));
	items.push_back (MenuElem (_("Ionian (\"Minor\")"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Ionian)));
	items.push_back (MenuElem (_("Phrygian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Phrygian)));
	items.push_back (MenuElem (_("Lydian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Lydian)));
	items.push_back (MenuElem (_("Mixolydian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Mixolydian)));
	items.push_back (MenuElem (_("Aeolian (\"Major\")"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Aeolian)));
	items.push_back (MenuElem (_("Locrian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Locrian)));
	items.push_back (MenuElem (_("Pentatonic Major"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::PentatonicMajor)));
	items.push_back (MenuElem (_("Pentatonic Minor"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::PentatonicMinor)));
	items.push_back (MenuElem (_("Major Chord"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::MajorChord)));
	items.push_back (MenuElem (_("Minor Chord"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::MinorChord)));
	items.push_back (MenuElem (_("Min7"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Min7)));
	items.push_back (MenuElem (_("Sus4"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Sus4)));
	items.push_back (MenuElem (_("Chromatic"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Chromatic)));
	items.push_back (MenuElem (_("Blues Scale"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::BluesScale)));
	items.push_back (MenuElem (_("Neapolitan Minor"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::NeapolitanMinor)));
	items.push_back (MenuElem (_("Neapolitan Major"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::NeapolitanMajor)));
	items.push_back (MenuElem (_("Oriental"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Oriental)));
	items.push_back (MenuElem (_("Double Harmonic"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::DoubleHarmonic)));
	items.push_back (MenuElem (_("Enigmatic"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Enigmatic)));
	items.push_back (MenuElem (_("Hirajoshi"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Hirajoshi)));
	items.push_back (MenuElem (_("Hungarian Minor"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::HungarianMinor)));
	items.push_back (MenuElem (_("Hungarian Major"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::HungarianMajor)));
	items.push_back (MenuElem (_("Kumoi"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Kumoi)));
	items.push_back (MenuElem (_("Iwato"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Iwato)));
	items.push_back (MenuElem (_("Hindu"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Hindu)));
	items.push_back (MenuElem (_("Spanish 8 Tone"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Spanish8Tone)));
	items.push_back (MenuElem (_("Pelog"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Pelog)));
	items.push_back (MenuElem (_("Hungarian Gypsy"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::HungarianGypsy)));
	items.push_back (MenuElem (_("Overtone"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Overtone)));
	items.push_back (MenuElem (_("Leading Whole Tone"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::LeadingWholeTone)));
	items.push_back (MenuElem (_("Arabian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Arabian)));
	items.push_back (MenuElem (_("Balinese"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Balinese)));
	items.push_back (MenuElem (_("Gypsy"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Gypsy)));
	items.push_back (MenuElem (_("Mohammedan"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Mohammedan)));
	items.push_back (MenuElem (_("Javanese"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Javanese)));
	items.push_back (MenuElem (_("Persian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Persian)));
	items.push_back (MenuElem (_("Algerian"), bind (mem_fun (*this, &SequenceUI::mode_fill), MusicalMode::Algerian)));
}

void
SequenceUI::build_trigger_menu ()

{
	using namespace Menu_Helpers;

	/* first, a submenu with the modes */
	
	Menu *mode_menu = manage (new Menu());
	MenuList& mode_items = mode_menu->items();
	RadioMenuItem::Group grp;

	mode_menu->set_name ("TriggerModeMenu");

	mode_items.push_back (RadioMenuElem (grp, "Fire once",
					     (mem_fun (*this, &SequenceUI::trigger_single_mode))));
	mode_items.push_back (RadioMenuElem (grp, "Fire many",
					     (mem_fun (*this, &SequenceUI::trigger_multiple_mode))));
	mode_items.push_back (RadioMenuElem (grp, "Restart when triggered",
					     (mem_fun (*this, &SequenceUI::trigger_multiretrig_mode))));
	mode_items.push_back (RadioMenuElem (grp, "Off when triggered",
					     (mem_fun (*this, &SequenceUI::trigger_off_mode))));

	/* Now the submenu with the list of triggers. The menu
	   itself is an object member.
	*/

	trigger_list_menu = new Menu();
	trigger_list_menu->set_name ("TriggerListMenu");
	
	/* next, the menu itself */

	trigger_menu = manage (new Menu());
	MenuList& trigger_menu_list = trigger_menu->items();

	trigger_menu->set_name ("TriggerSelectorMenu");

	trigger_menu_list.push_back (MenuElem ("Mode", *mode_menu));
	trigger_menu_list.push_back (MenuElem ("Use", *trigger_list_menu));

	/* rebuild the menu of available triggers every time we
	   map the trigger menu.
	*/

	trigger_menu->signal_map_event().connect
		(mem_fun (*this, &SequenceUI::rebuild_trigger_menu));
}

bool
SequenceUI::rebuild_trigger_menu (GdkEventAny *ignored)

{
	using namespace Menu_Helpers;
	MenuList& trigger_list = trigger_list_menu->items ();
	MenuList::iterator i;

	trigger_list.clear();

	/* add the special entry */

	trigger_list.push_back (CheckMenuElem ("None"));
	
	/* make sure its marked correctly */

	dynamic_cast<CheckMenuItem*>(&trigger_list.back())->set_active (_sequence.get_trigger_name().empty());

	trigger_list.back().signal_activate().connect 
		(mem_fun (*this, &SequenceUI::trigger_dropped));

	/* now add the rest */
	
	_sequence.sequencer().foreach_trigger 
		(*this, &SequenceUI::add_trigger_item);

	return false;
}

int
SequenceUI::add_trigger_item (Trigger &trigger)

{
	using namespace Menu_Helpers;

	MenuList& trigger_list_items = trigger_list_menu->items();

	trigger_list_items.push_back (CheckMenuElem (trigger.name()));
	
	if (trigger.name() == _sequence.get_trigger_name()) {
		dynamic_cast<CheckMenuItem*>(&trigger_list_items.back())->set_active (true);
	}

	trigger_list_items.back().signal_activate().connect (bind (mem_fun (*this, &SequenceUI::trigger_selected), &trigger));

	return 0;
}

int
SequenceUI::delete_trigger_item (Trigger &trigger)

{
	return 0;
}

void
SequenceUI::trigger_dropped ()

{
	_sequence.drop_trigger ();
}

void
SequenceUI::trigger_selected (Trigger *trigger)
{
	if (trigger->name() == "None") {
		_sequence.drop_trigger ();
	} else {
		_sequence.use_trigger (*trigger);
	}
}

void
SequenceUI::trigger_single_mode ()

{
	_sequence.set_trigger_mode (Sequence::single);
}

void
SequenceUI::trigger_multiple_mode ()

{
	_sequence.set_trigger_mode (Sequence::multiple);
}

void
SequenceUI::trigger_multiretrig_mode ()

{
	_sequence.set_trigger_mode (Sequence::multretrig);
}

void
SequenceUI::trigger_off_mode ()

{
	_sequence.set_trigger_mode (Sequence::off_trigger);
}

void
SequenceUI::mode_fill (MusicalMode::Type t)
{
	/* number of octaves is 0..5, random */
	_sequence.mode_fill (t, random()%6);
}

void
SequenceUI::trigger_click ()
{
	if (!trigger_menu) {
		build_trigger_menu ();
	}
	trigger_menu->popup (1, 0);
}

void
SequenceUI::modefill_click ()
{
	if (!modefill_menu) {
		build_modefill_menu ();
	}
	modefill_menu->popup (1, 0);
}
