/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: sequencer.cc,v 1.9 2000/11/25 00:10:31 pbd Exp $
*/

#include <cstring>

#include <stdio.h>
#include <sys/time.h>
#include <signal.h>
#include <sched.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <poll.h>
#ifndef __apple__
#include <sys/mman.h>
#endif

#include <sigc++/bind.h>

#include <pbd/error.h>
#include <pbd/failed_constructor.h>
#include <pbd/pthread_utils.h>
#include <pbd/compose.h>
#include <pbd/stacktrace.h>

#include <midi++/port.h>
#include <midi++/channel.h>
#include <midi++/parser.h>
#include <midi++/manager.h>

#include "step.h"
#include "sequence.h"
#include "sequencer.h"
#include "trigger.h"
#include "patternfile.h"
#include "configuration.h"

#include "i18n.h"

#include <pbd/stl_delete.h>

#ifdef USE_RTC
#include <linux/rtc.h>
#include <sys/ioctl.h>
#endif /* USE_RTC */

#include <pbd/stl_delete.h>

#define CLOCK_INTERVAL 10000 /* nominal usecs per tick */

using namespace PBD;
using namespace sigc;

Sequencer *Sequencer::theSequencer = 0;

int
Sequencer::midi_read (MIDI::Port& port)
{
	MIDI::byte buf[512];
	
	/* reading from the MIDI port activates the Parser
	   that in turn generates signals that we care
	   about. the port is already set to NONBLOCK so that
	   can read freely here.
	*/
	
	while (1) {
		
		// cerr << "+++ READ ON " << port.name() << endl;
		
		int nread = port.read (buf, sizeof (buf));

		// cerr << "-- READ (" << nread << " ON " << port.name() << endl;
		
		if (nread > 0) {
			if ((size_t) nread < sizeof (buf)) {
				break;
			} else {
				continue;
			}
		} else if (nread == 0) {
			break;
		} else if (errno == EAGAIN) {
			break;
		} else {
			fatal << "Error reading from MIDI port " << port.name() << endmsg;
			/*NOTREACHED*/
		}
	}

	return 0;
}

void *
Sequencer::_sequencer_thread (void *arg)
{
	return static_cast<Sequencer*> (arg)->sequencer_thread ();
}

void *
Sequencer::sequencer_thread ()
{
	struct pollfd* pfd;
	int nfds = 0;
	int pfd_size = 0;
	struct sched_param rtparam;
	list<MIDI::Port*> ports;
	ticks_t next_wakeup = TICK_MAX;

	PBD::ThreadCreated (pthread_self(), "sequencer");

	if (Config->use_real_time) {

#ifndef __apple__
	
		memset (&rtparam, 0, sizeof (rtparam));
		rtparam.sched_priority = 9;
		
		if (pthread_setschedparam (pthread_self(), SCHED_FIFO, &rtparam) != 0) {
			warning << _("Not able to run with realtime scheduling") << endmsg;
		} else {
			info << _("Running with realtime scheduling") << endmsg;
			if (mlockall (MCL_CURRENT | MCL_FUTURE) != 0) {
				warning << string_compose (_("Error locking memory (might affect realtime performace): %1"), strerror (errno)) << endmsg;
			}

		}
#endif

	}

	const MIDI::Manager::PortMap& p = MIDI::Manager::instance()->get_midi_ports ();

	for (MIDI::Manager::PortMap::const_iterator i = p.begin(); i != p.end(); ++i) {
		ports.push_back ((*i).second);
	}
	
	pfd_size = ports.size() + 1;
	pfd = new pollfd[pfd_size];

	while (1) {

		int n;
		
		list<MIDI::Port*>::iterator i;

		/* Always poll the thread pipe */

		pfd[0].fd = thread_pipe[0];
		pfd[0].events = POLLIN|POLLHUP|POLLERR;
		nfds = 1;

		if (thread_running) {
			
			/* only poll ports if we're meant to be busy */

			for (i = ports.begin(), n = 1; i != ports.end() && n < pfd_size; ++n, ++i) {
				pfd[n].fd = (*i)->selectable();
				pfd[n].events = POLLIN|POLLHUP|POLLERR;
				nfds++;
			}
		} 

		int nready;
		if ((nready = poll (pfd, nfds, poll_timeout)) < 0) {
			if (errno == EINTR) {
				/* gdb at work, perhaps */
				continue;
			}

			error << "MIDI thread poll failed (" 
			      << strerror (errno)
			      << endmsg;

			break;
		}

		if (nready == 0) {
			goto tick;
		}

		if (pfd[0].revents & ~POLLIN) {
			break;
		}
		
		if (pfd[0].revents & POLLIN) {
			char buf[64];
			bool finished = false;

			while (!finished) {
				int x = read (thread_pipe[0], buf, sizeof (buf));
				if (x < 0) {
					if (errno == EAGAIN) {
						finished = true;
						continue;
					}

					/* else error on pipe */

					error << "read error on thread pipe: "
					      << strerror (errno)
					      << endmsg;
					return 0;

				} else if (x == 0) {
					finished = true;
				}
			}


			if (state == running) {
				poll_timeout = CLOCK_INTERVAL/1000; // usecs->millisecs
			} else {
				poll_timeout = -1;
			}
			continue;
		}

		for (i = ports.begin(), n = 1; i != ports.end() && n < nfds; ++n) {

			list<MIDI::Port*>::iterator tmp;

			tmp = i;
			++tmp;

			if (pfd[n].revents & ~POLLIN) {
				ports.erase (i);
			} else if (pfd[n].revents & POLLIN) {
				midi_read (**i);
			}
			
			i = tmp;
		}


	  tick:
		next_wakeup = clock_tick ();
		
		if (state == running) {

			/* only adjust next wakeup if we're running */

			if (next_wakeup != TICK_MAX) {
				if (next_wakeup == 0) {
					/* failed to take the lock, "spin" and try again */
					poll_timeout = 1; // msec
				} else {
					poll_timeout = (next_wakeup - midi_ticks) * usecs_per_tick/1000;
					if (poll_timeout > 100) {
						poll_timeout = 100;
					}
				} 
			}

		} 
	}

	return 0;
}

Sequencer::Sequencer (size_t num_steps, size_t num_seqs, bool with_rtc)
{
	use_rtc = with_rtc;
	thread_running = false;
	poll_timeout = -1;

	if (pipe (thread_pipe)) {
		error << "could not create thread pipe ("
		      << strerror (errno)
		      << ')'
		      << endmsg;
		throw failed_constructor();
	}
	
	if (fcntl (thread_pipe[0], F_SETFL, O_NONBLOCK)) {
		error << "Sequencer: cannot set O_NONBLOCK on thread read pipe ("
		      << strerror (errno)
		      << ')'
		      << endmsg;
		throw failed_constructor();
	}

	if (fcntl (thread_pipe[1], F_SETFL, O_NONBLOCK)) {
		error << "Sequencer: cannot set O_NONBLOCK on thread write pipe ("
		      << strerror (errno)
		      << ')'
		      << endmsg;
		throw failed_constructor();
	}

	if (pthread_create (&thread_id, NULL, _sequencer_thread, this)) {
		throw failed_constructor();
	}

	if (theSequencer == 0) {
		theSequencer = this;
	}

	carry_usecs = 0;
	last_tick_usecs = 0;

	default_steps = num_steps;

	for (size_t i = 0; i < num_seqs; i++) {
		new_sequence ();
	}

	reset();
}

Sequencer::~Sequencer ()
{
	pthread_cancel (thread_id);
	list_delete (&sequences);
}

void
Sequencer::reset ()
{
	Sequences::iterator s;
	Glib::Mutex::Lock lm (_lock);

	state = stopped;
	set_clock (false);
	midi_ticks = 0;
	
	/* use 120BPM as the initial value */

	usecs_per_tick = (60000000 / 120) / Config->ticks_per_beat;

	start_usecs = 0;

	for (s = sequences.begin(); s != sequences.end(); s++) {
		(*s)->reset ();

		for (size_t r = 0; r < MAX_CHANNELS; r++) {
			(*s)->write_port()->channel(r)->all_notes_off ();
		}

		/* XXX reset controllers */
	}

	set_delete (&_triggers);
	_copy_sequence = -1;
}

Sequence *
Sequencer::new_sequence (bool notify)
{
	Sequence *seq;

	{
		Glib::Mutex::Lock lm (_lock);

		seq = new Sequence (sequences.size(), 
				    default_steps,
				    *this);
		sequences.push_back (seq);
	}

	if (notify) {
		sequence_added (*seq);
	}

	return seq;
}

void 
Sequencer::start ()
{
	srand (time((time_t)0));
	run (); /* start thread */
	state = running;
	set_clock (true);
}

void 
Sequencer::halt ()
{
	state = stopped;
	set_clock (false);
	stop (); /* stop thread */
	reset ();
}

void
Sequencer::pause_clock ()
{
	if (state != running) {
		return;
	}
	state = paused;
	set_clock (false);
}

void
Sequencer::play ()
{
	if (state != running) {
		resume_clock ();
	}
}

void
Sequencer::pause ()
{
	if (state == running) {
		pause_clock ();
	}
}

void
Sequencer::toggle_pause ()
{
	if (state != running) {
		resume_clock ();
	} else {
		pause_clock ();
	}
}

void
Sequencer::resume_clock()
{
	if (state != paused && state != stopped) {
		return;
	}
	state = running;
	set_clock (true);
}

void
Sequencer::sync ()
{
	Sequences::iterator s;
	unsigned int now = midi_ticks;
	Glib::Mutex::Lock lm (_lock);

	for (s = sequences.begin(); s != sequences.end(); s++) {
		(*s)->set_current_step ((*s)->first_step() + 
					(*s)->start_step());
		(*s)->set_next_gate (now + (*s)->gate_interval());
	}
}

void
Sequencer::compute_time ()
{
	struct timeval tcurrent_tick;
	struct timezone tz;
	unsigned long current_usecs;
	unsigned long elapsed_usecs;
	ticks_t inc;

	/* If you can come up with a simpler way to provide the same
	   reliability as thus tricky mess of code, I'd love to hear about
	   it. Several times, I thought I had it, but immediately found
	   common situations where simpler code just wouldn't work right.
	*/

	gettimeofday (&tcurrent_tick, &tz);

	current_usecs = (tcurrent_tick.tv_sec * 1000000) + tcurrent_tick.tv_usec;
	elapsed_usecs = carry_usecs + (current_usecs - last_tick_usecs);

	if (usecs_per_tick > CLOCK_INTERVAL) {
	
		/* Each timer signal does not necessarily correspond to one
		   tick. So, check whether the cumulative elapsed
		   time is enough, we have a new tick.
		*/
	
		if (elapsed_usecs < usecs_per_tick) {
			inc = 0;
		} else if (elapsed_usecs > usecs_per_tick) {
			inc = (ticks_t) ((double) elapsed_usecs /
					 (double) usecs_per_tick);
		} else {
			inc = 1;
		}

	} else { 

		/* every timer signal corresponds to at least one
		   tick. Found out how much time has elapsed, and
		   figure out just how many ticks have gone by. But
		   note that the elapsed time might not be greater
		   than usecs_per_tick, so we have to allow some slop
		   here.  
		*/

		if (elapsed_usecs >= (usecs_per_tick*2)) {
			inc = (ticks_t) ((double) elapsed_usecs /
					 (double) usecs_per_tick);
		} else {
			inc = 1;
	    
		}

	}

	if (inc > 0) {
		midi_ticks += inc;
		last_tick_usecs = current_usecs;
		carry_usecs = elapsed_usecs - (inc * usecs_per_tick);
		carry_usecs = carry_usecs > 0 ? carry_usecs : 0;
	}
}

void
Sequencer::foreach_sequence (sigc::slot<void,Sequence&> theslot)
{
	Sequences::iterator s;

	for (s = sequences.begin(); s != sequences.end(); s++) {
		theslot (**s);
	}
}

void
Sequencer::advance_sequence (Sequence &seq, unsigned int now)
{
	/* now == 0 => first execution - no step advance */
	
	SequenceLocker sl (&seq);

	if (!sl.locked() || now == 0 || !seq.needs_attention (now)) {
		return;
	}
	
	seq.advance ();
}

void
Sequencer::check_sequence (Sequence &seq, unsigned int now)
{
	SequenceLocker sl (&seq);

	if (!sl.locked()) {
		return;
	}

	seq.check (now);
}

void
Sequencer::execute_sequence (Sequence &seq, unsigned int now)
{
	SequenceLocker sl (&seq);

	if (!sl.locked()) {
		return;
	}

	seq.execute (now);
}

void
Sequencer::get_next_wakeup (Sequence& seq, ticks_t& current_wakeup)
{
	seq.get_next_wakeup (current_wakeup);
}

ticks_t
Sequencer::clock_tick ()
{	
	ticks_t wakeup = TICK_MAX;

	if (!_lock.trylock()) {
		cerr << "MIDI clock locked out" << endl;
		wakeup = 0;
		return wakeup; /* we'll be back momentarily ... */
	}

	compute_time ();
	foreach_sequence (bind (sigc::mem_fun (*this, &Sequencer::check_sequence), midi_ticks));
	foreach_sequence (bind (sigc::mem_fun (*this, &Sequencer::advance_sequence), midi_ticks));
	foreach_sequence (bind (sigc::mem_fun (*this, &Sequencer::execute_sequence), midi_ticks));

	/* XXX bind() doesn't work correctly with reference arguments */

	for (Sequences::iterator s = sequences.begin(); s != sequences.end(); ++s) {
		(*s)->get_next_wakeup (wakeup);
	}

	tick (midi_ticks);

	_lock.unlock ();

	return wakeup;
}

void
Sequencer::emit_clock ()
{
	Sequences::iterator s;

	for (s = sequences.begin(); s != sequences.end(); s++) {
		(*s)->write_port()->clock();
	}
}

unsigned int
Sequencer::set_clock (bool running)
{
	/* break into poll loop. it will wake up and reset poll_timeout appropriately */
	
	char c = 0;
	write (thread_pipe[1], &c, 1);
	
	return 0;
}

void
Sequencer::set_base_note (MIDI::Port *p, int channel, int note)
{
	Sequences::iterator s;

	for (s = sequences.begin(); s != sequences.end(); s++) {

		if ((*s)->read_port() == p &&
		    (*s)->input_channel() == channel) {
			(*s)->set_base_note (note, 0);
		}
	}

}

int 
Sequencer::sumover (Sequence::Mode mode, int row)

{
	int set = -1;
	int val = 0;
	Step *step;
	Sequences::iterator s;

	for (s = sequences.begin(); s != sequences.end(); s++) {

		if ((*s)->mode() == mode && (*s)->controls_row (row)) {
	    
			step = (*s)->step ((*s)->current_step());
	    
			if (step->is_enabled()) {
		
				val += step->val();
		
				if (mode == Sequence::note_on_velocity ||
				    mode == Sequence::note_off_velocity) {
					val = val > 0x7F ? 0x7F : val;
				} 
		
				set = 1;
			}
		}
	}

	return set > 0 ? val : -1;
}

void
Sequencer::cancel_all_triggers ()
{
	TriggerMap::iterator t;

	for (t = _triggers.begin(); t != _triggers.end(); t++) {
		(*t)->reset ();
	}
}

void
Sequencer::dump (std::ostream &o)
{
	Sequences::iterator s;
	TriggerMap::iterator t;

	pause_clock ();

	o << usecs_per_tick << ' '
	  << sequences.size() << ' '
	  << _triggers.size() << ' ';
	
	for (s = sequences.begin(); s != sequences.end(); s++) {
		o << **s;
	}

	for (t = _triggers.begin(); t != _triggers.end(); t++) {
		o << **t;
	}

	resume_clock();
}

void
Sequencer::load (std::istream &str)
{
	size_t nsequences;
	size_t ntriggers;
	Sequences::iterator siter;
	TriggerMap::iterator titer;

	pause_clock ();

	begin_pattern_loading ();

	str >> usecs_per_tick;
	str >> nsequences;
	str >> ntriggers;

	/* triggers come first so that sequences can hook up with
	   them by name 
	*/

	set_delete (&_triggers);

	for (titer = _triggers.begin(); ntriggers > 0; ntriggers--, titer++) {
		Trigger *t = new Trigger;

		str >> *t;
		_triggers.insert (t);
	}

	while (sequences.size() < nsequences) {
		new_sequence (true);
	}
    
	for (siter = sequences.begin(); nsequences-- ; siter++) {
		str >> **siter;
	}

	resume_clock ();
	reloaded ();
}

void
Sequencer::set_copy_sequence (int copy)
{
	bool send = (copy != _copy_sequence);

	if (copy > (int) sequences.size()) {
		return;
	}

	_copy_sequence = copy;
	
	if (send) {
		copy_sequence_changed (_copy_sequence);
	}
}

void
Sequencer::unset_copy_sequence (int copy)
{
	if (_copy_sequence == copy) {
		_copy_sequence = -1;
		copy_sequence_changed (_copy_sequence);
	}
}

Sequence*
Sequencer::nth_sequence(int n)
{
	Sequences::iterator s;

	if (n < 0) {
		return 0;
	}

	s = sequences.begin();
	while (s != sequences.end() && n) {
		--n;
		++s;
	}

	if (s == sequences.end() || n) {
		return 0;
	}

	return *s;
}

void
Sequencer::paste_sequence (int dest)
{
	Sequence* from = nth_sequence (_copy_sequence);
	Sequence* to = nth_sequence (dest);

	if (to && from) {
		to->paste_from (*from);
	}
}

std::ostream& 
operator<< (std::ostream& ostr, Sequencer &seq)
{
	seq.dump (ostr);
	return ostr;
}

std::istream& 
operator>> (std::istream& istr, Sequencer &seq) 
{
	seq.load (istr);
	return istr;
}

void
Sequencer::midi_clock_signal (MIDI::Parser &p)
{
	static struct timeval then = { 0, 0 };
	struct timeval now;

	gettimeofday (&now, 0);

	if (then.tv_usec == 0) { 
		then = now;
		return;
	}

	/* May need some kind of hysteresis on this - many external
	   sources vary by substantial amounts even when their own
	   notion of time is constant.
	*/

	
	usecs_per_tick = ((now.tv_sec * 1000000) + now.tv_usec) -
		((then.tv_sec * 1000000) + then.tv_usec);

	then = now;
}

void
Sequencer::midi_stop_signal (MIDI::Parser &p)

{
	Sequences::iterator i;

	pause_clock ();

	/* reset every sequence so that it start immediately
	   on continue/start ...
	*/
	for (i = sequences.begin(); i != sequences.end(); i++) {
		(*i)->set_next_gate (0);
	}
}

void
Sequencer::midi_start_signal (MIDI::Parser &p)
{
	resume_clock ();
}

void
Sequencer::midi_continue_signal (MIDI::Parser &p)
{
	if (state != running) {
		clock_tick ();
	}

	resume_clock ();
}


bool
Sequencer::add_trigger (Trigger &trig)
{
	std::pair<TriggerMap::iterator, bool> result;

	result = _triggers.insert (&trig);

	if (result.second) { 
		/* new trigger */
		trig.deleted.connect (mem_fun (*this, &Sequencer::trigger_deleted));
	}

	return result.second;
}

void
Sequencer::trigger_deleted (Trigger &trig)
{
	TriggerMap::iterator x;

	if ((x = _triggers.find (&trig)) != _triggers.end()) {
		_triggers.erase (x);
	} 
}

Trigger *
Sequencer::trigger (const string &name)
{
	TriggerMap::iterator i;

	for (i = _triggers.begin(); i != _triggers.end(); i++) {
		if ((*i)->name() == name) {
			return (*i);
		}
	}
	return 0;
}

void
Sequencer::use_MIDI_clock (bool yn)
{
	if (yn) {

		clock_connection = 
		MIDI::Manager::instance()->inputPort->input()-> timing.connect
			(mem_fun (*this, &Sequencer::midi_clock_signal));

		start_connection = 
		MIDI::Manager::instance()->inputPort->input()->start.connect 
			(mem_fun (*this, &Sequencer::midi_start_signal));

		stop_connection = 
		MIDI::Manager::instance()->inputPort->input()->stop.connect 
			(mem_fun (*this, &Sequencer::midi_stop_signal));

		continue_connection = 
		MIDI::Manager::instance()->inputPort->input()->contineu.connect 
			(mem_fun (*this, &Sequencer::midi_continue_signal));



	} else {
                /* these are harmless if not connected */

		clock_connection.disconnect(); 
		stop_connection.disconnect ();
		start_connection.disconnect ();
		continue_connection.disconnect ();
	}
}

void
Sequencer::generate_MIDI_clock (bool yn)

{
	/* XXX not sure how to do this right now */
}

void
Sequencer::run ()
{
	char c = 1; /* contents don't matter */
	thread_running = true;
	write (thread_pipe[1], &c, 1);
}

void
Sequencer::stop ()
{
	char c = 1; /* contents don't matter */
	thread_running = false;
	write (thread_pipe[1], &c, 1);
}
