/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: sequencer.h,v 1.6 2000/11/25 00:10:31 pbd Exp $
*/

#ifndef __sw_sequencer_h__
#define __sw_sequencer_h__

#include <list>
#include <set>
#include <pthread.h>
#include <glibmm/thread.h>
#include <sigc++/signal.h>
#include <pbd/error.h>

#include "types.h"
#include "sequence.h"
#include "trigger.h"

class Step;
class PatternFile;

namespace MIDI {
	class Port;
}

/* defaults for when nothing is set ... */

#define DEFAULT_OFF_VELOCITY 126
#define DEFAULT_BASE_NOTE 60
#define MAX_CHANNELS 16


class Sequencer : public sigc::trackable 
{
  public:
	enum SequencerState { running, paused, stopped };
	Sequencer (size_t rows, size_t steps, bool with_rtc);
	~Sequencer();

	void start ();       /* start clock, action, etc. */
	void halt ();        /* stop everything, reset */
	void play (); 
	void pause (); 
	void toggle_pause (); 
	void sync ();        /* return all rows to start step */
	void reset ();       /* return entire state to default */

	SequencerState get_state() const { return state; }

	void foreach_sequence (sigc::slot<void,Sequence&>);

	void set_patternfile (PatternFile *pf);

	ticks_t ticks() { return midi_ticks; }
	void set_usecs_per_tick (unsigned long usecs) {
		Glib::Mutex::Lock lm (_lock);
		usecs_per_tick = usecs;
	}

	unsigned int current_usecs_per_tick () { return usecs_per_tick; }

	/* Step manipulation */

	void toggle_step (int step, int on);

	/* These two look like they should be row-level access functions,
	   but in fact they are a property of the entire state of the
	   sequencer.
	*/

	MIDI::byte on_velocity (int row);
	MIDI::byte off_velocity (int row);

	int sumover (Sequence::Mode, int for_seq);

	/* Trigger manipulation */

	Trigger  *trigger        (const std::string &);
	bool      add_trigger    (Trigger &t);
	size_t    ntriggers() const { return _triggers.size(); }

	template<class T> int foreach_trigger (T &obj, int (T::*)(Trigger &));

	/* use/sending MIDI clock messages */

	void generate_MIDI_clock (bool);
	void use_MIDI_clock (bool);

	sigc::signal1<void,unsigned int> clock_change;
	sigc::signal0<void> reloaded;
	sigc::signal1<void, size_t> tick;
	sigc::signal0<void> begin_pattern_loading;
	sigc::signal0<void> end_pattern_loading;
	sigc::signal1<void,int> copy_sequence_changed;
	sigc::signal1<void,Sequence &> sequence_added;

	static Sequencer *instance () { return theSequencer; }

  protected:
	friend class SequencerUI;
	Sequence *new_sequence (bool notify = false);

	friend class SequenceUI;
	void set_copy_sequence (int copy);
	void unset_copy_sequence (int copy);
	void paste_sequence (int destination);
	int  copy_sequence () { return _copy_sequence; }

	friend std::ostream& operator<< (std::ostream&, Sequencer &);
	friend std::istream& operator>> (std::istream&, Sequencer &);
	void load (std::istream &);
	void dump (std::ostream &);

  private:
	pthread_t thread_id;
	bool use_rtc;
	int _copy_sequence;
	unsigned int default_steps;
	ticks_t  midi_ticks;       /* ticks since last reset */
	unsigned long last_tick_usecs;
	unsigned long carry_usecs;
	unsigned long usecs_per_tick;  
	unsigned long start_usecs;     
	int poll_timeout;
	int rtc_fd;

	void pause_clock ();       /* stop wherever we are right now, no reset */
	void resume_clock ();      /* continue after pause() */

	/* Signal connections */

	sigc::connection raw_connection;

	/* Sequences */
	
	class SequenceLocker {
	  public:
	    SequenceLocker (Sequence* s) {
		    seq = s;
		    ok = s->acquire_lock ();
	    }
	    ~SequenceLocker() {
		    if (ok) {
			    seq->release_lock ();
		    }
	    }
	    bool locked () const { return ok; }

	  private:
	    Sequence* seq;
	    bool ok;
	};

	typedef std::list<Sequence *> Sequences;
	Sequences sequences;
	Sequence* nth_sequence (int);

	/* Triggers */

	typedef std::set<Trigger *> TriggerMap;
	std::set<Trigger *> _triggers;
	void cancel_all_triggers ();
	void trigger_deleted (Trigger &);

	SequencerState state;
	
	unsigned int set_clock (bool onoff);

	ticks_t clock_tick ();
	void compute_time ();

	Glib::Mutex _lock;

	/* handling external MIDI clock messages */

	sigc::connection clock_connection;
	sigc::connection start_connection;
	sigc::connection stop_connection;
	sigc::connection continue_connection;

	void midi_clock_signal (MIDI::Parser &);
	void midi_start_signal (MIDI::Parser &);
	void midi_stop_signal (MIDI::Parser &);
	void midi_continue_signal (MIDI::Parser &);

	void emit_clock ();
	
	void check_sequence (Sequence &, unsigned int);
	void advance_sequence (Sequence &, unsigned int);
	void execute_sequence (Sequence &, unsigned int);
	void get_next_wakeup (Sequence &, ticks_t&);

	void set_base_note (MIDI::Port *p, int channel, int note);
	void handle_note_on (MIDI::Parser &, MIDI::EventTwoBytes *);

	/* Thread activity */

	static void* _sequencer_thread (void*);
	void* sequencer_thread ();
	int midi_read (MIDI::Port&);
	int thread_pipe[2];
	bool thread_running;
	void stop ();
	void run ();

	static Sequencer *theSequencer;

};

std::ostream& operator<< (std::ostream&, Sequencer &);
std::istream& operator>> (std::istream&, Sequencer &);

template<class T> int
Sequencer::foreach_trigger (T &obj, int (T::* meth)(Trigger &))

{
	TriggerMap::iterator i;
	
	for (i = _triggers.begin(); i != _triggers.end(); i++) {
		if ((obj.*meth) (**i)) {
			return -1;
		}
	}
	
	return 0;
}
 

#endif /*  __sw_sequencer_h__ */


