/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: step.cc,v 1.3 2000/05/03 15:54:21 pbd Exp $
*/

#include <stdio.h>
#include <strings.h>
#include <sys/time.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <pbd/error.h>

#include "step.h"
#include "sequence.h"
#include "sequencer.h"

using namespace MIDI;
using namespace PBD;

Step::Step (size_t i, Sequence &s)
	: MIDI::Controllable (s.read_port(), s.input_channel()),
	  _sequence (s),
	  _id (i)
{
	ptr = NULL;
	reset ();
	
	_sequence.input_midi_changed.connect (mem_fun (*this, &Step::input_midi_changed));
}

void
Step::reset ()
{
	if (ptr) {
		delete [] ptr;
	}

	ptr = NULL;
	num = DEFAULT_BASE_NOTE;
	time_on = 0;
	time_off = 0;
	enabled = true;
	on = false;
	_is_start = false;
	_is_end = false;
	_is_current = false;
}

ticks_t
Step::get_next_wakeup () const
{
	if (_sequence.makes_sound()) {
		if (on) {
			return time_off;
		} else {
			return TICK_MAX;
		}
	} else {
		return TICK_MAX;
	}
}

void
Step::note_off ()

{
	_sequence.note_off (num);
	on = false;
	time_on = 0;
	time_off = 0;
}

void
Step::note_on (ticks_t when, ticks_t duration)

{
	on = true;
	time_on = when;
	time_off = when + duration;
	_sequence.note_on (num);
}

void
Step::disable ()

{
	bool send = enabled;

	if (on) { /* true only for noteOn/noteOff steps */
		note_off();
	}

	enabled = false;

	if (send) {
		active_change ();
	}
}

void
Step::enable ()

{
	bool send = !enabled;

	enabled = true;

	if (send) {
		active_change ();
	}
}

void
Step::set (int n, bool in_single_pitch_set)
{
	if (num == (byte) n && !in_single_pitch_set) {
		return;
	}

	if (_sequence.is_raw_midi()) {
		fatal << "numeric step set with row as midi" << endmsg;
	}
	
	if (on) {
		note_off ();
	}
	
	if (n < 0) n = 127;
	if (n > 127) n = 0;

	num = (byte) n;
	
	if (!in_single_pitch_set && 
	    _sequence.mode() == Sequence::single_pitch) {
		_sequence.set_all_steps (n);
	} 
	
	value_change ();
}

void
Step::set (byte *msg, size_t size)

{
	if (!_sequence.is_raw_midi()) {
		fatal << "midibyte step set with row not as midi" << endmsg;
	}

	if (ptr && ptr != msg) {
		delete [] ptr;
	}
    
	num = size;
	ptr = msg;

	if (ptr == 0 || num == 0) {
		disable ();
	}

	value_change ();
}

void
Step::check (unsigned int now)

{
	/* XXX will this be clean enough for chords ?
	   How much possible time can elapse between
	   the first note off and the last for the chord ?
	*/

	if (_sequence.makes_sound()) {
		if (on && time_off <= now) {
			note_off();
		}
	}
}

void
Step::execute (unsigned int now)

{
	if (!enabled || _sequence.off()) {
		return;
	}

	switch (_sequence.mode()) {
	case Sequence::absolute_pitch:
	case Sequence::relative_pitch:
	case Sequence::single_pitch:
		if (!_sequence.silent()) {
			note_on (now, _sequence.note_duration ());
		}
		break;
	  
	case Sequence::controller:
		_sequence.control (_sequence.output_channel(),
				    _sequence.controller_id(), num);
		break;
	  
	case Sequence::recorded_midibytes:
	case Sequence::midibytes:
		if (num > 0) {
			_sequence.midimsg (ptr, num);
		}
		break;

	case Sequence::gate_timing:
	case Sequence::note_length:
	case Sequence::note_on_velocity:
	case Sequence::note_off_velocity:
		/* nothing to do */
		break;

	case Sequence::program_change:
		_sequence.channel_msg (_sequence.output_channel(),
					MIDI::program,
					num, 0);
		break;
	case Sequence::pitch_bend:
		_sequence.channel_msg (_sequence.output_channel(),
					MIDI::pitchbend,
					num, 0);
		break;
	case Sequence::polytouch:
		_sequence.channel_msg (_sequence.output_channel(),
					MIDI::polypress,
					num, 0);
		break;
	case Sequence::aftertouch:
		_sequence.channel_msg (_sequence.output_channel(),
					MIDI::chanpress,
					num, 0);
		break;

	case Sequence::absolute_chord:
	case Sequence::relative_chord:
		/* handled at _sequence level */
		break;

	default:
		fatal << "Step::execute: this can't happen" << endmsg;

	}
}

void
Step::dump (std::ostream &str)

{
	str << (enabled ? 1 : 0) << ' '
	    << num << ' '
	    << (ptr ? 1 : 0) << ' ';

	if (ptr) {
		for (size_t i = 0; i < num; i++) {
			str << (int) ptr[i] << ' ';
		}
	} 
}

void
Step::load (std::istream &str)

{
	int load_ptr;
	unsigned int newval;

	reset();

	str >> newval; enabled = (bool) newval;
	str >> newval;
	str >> load_ptr;

	if (load_ptr) {
		ptr = new byte [newval];
		for (size_t i = 0; i < newval; i++) {
			int n;
			str >> n;
			ptr[i] = (byte) n;
		}
		set (ptr, newval);
	} else {
		set (newval);
	}
}

void
Step::set_current (bool is_current)

{
	bool send = (_is_current != is_current);

	_is_current = is_current;

	if (send) {
		current_change ();
	}
}

void
Step::set_end (bool is_end)

{
	bool send = (_is_end != is_end);

	_is_end = is_end;

	if (send) {
		terminal_change ();
	}
}

void
Step::set_start (bool is_start)

{
	bool send = (_is_start != is_start);

	_is_start = is_start;

	if (send) {
		terminal_change ();
	}
}

void
Step::input_midi_changed ()

{
	drop_external_control ();
	midi_rebind (_sequence.read_port(), _sequence.input_channel());
}

/* ---  MidiControllable interface ------------------------------- */

void
Step::set_value (float val)

{
	if (!_sequence.is_raw_midi ()) {
		set ((int) val);
	}
}
	
float
Step::lower ()

{
	return _sequence.minimum_step_value();
}

float
Step::upper ()

{
	return _sequence.maximum_step_value ();
}

float
Step::range ()

{
	return upper() - lower() + 1;
}

std::ostream &
operator<< (std::ostream &ostr, Step &step) 
{
	step.dump (ostr);
	return ostr;
}

std::istream &
operator>> (std::istream &istr, Step &step) 

{
	step.load (istr);
	return istr;
}
