/*
    Copyright (C) 1999-2007 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: step_ui.cc,v 1.7 2000/11/25 00:10:31 pbd Exp $
*/

#include <algorithm>

#include <gtkmm2ext/gtk_ui.h>
#include <gtkmm2ext/utils.h>

#include "ui.h"
#include "step_ui.h"
#include "step.h"
#include "midibyte_edit.h"

#include "i18n.h"

StepUI::StepUI (Step &s) 
	: step (s),
	  adjustment (0, s.lower(), s.upper(), 1, 4),
	  learning_popup (Gtk::WIN_POS_MOUSE, 30000, false)
{
	terminalHBox.pack_start (terminalButton, false, false, 10);
	pack_start (terminalHBox, false, false, 10);

	valueDisplay.set_name ("StepValueDisplay");

	Gtkmm2ext::set_size_request_to_display_given_text (valueDisplay, "127", 10, 10);

	step.learning_started.connect (mem_fun (*this, &StepUI::learning_started));
	step.learning_stopped.connect (mem_fun (*this, &StepUI::learning_stopped));

	valueDisplay.add_events (Gdk::BUTTON_RELEASE_MASK|
				 Gdk::ENTER_NOTIFY_MASK|
				 Gdk::LEAVE_NOTIFY_MASK|
				 Gdk::SCROLL_MASK);
	
	button_release_can_toggle = false;

	valueDisplay.set_has_frame (false);
	valueDisplay.signal_focus_in_event().connect (mem_fun (*this,&StepUI::focus_in_step));
	valueDisplay.signal_focus_out_event().connect (mem_fun (*this,&StepUI::focus_out_step));
	valueDisplay.signal_leave_notify_event().connect (mem_fun (*this,&StepUI::pointer_left_step));
	valueDisplay.signal_leave_notify_event().connect (mem_fun (*this,&StepUI::pointer_left_step));
	valueDisplay.signal_enter_notify_event().connect (mem_fun (*this,&StepUI::pointer_entered_step));
	valueDisplay.signal_scroll_event().connect (mem_fun (*this,&StepUI::value_display_scroll_event));
	valueDisplay.signal_button_press_event().connect (mem_fun (*this,&StepUI::value_display_button_press), false);
	valueDisplay.signal_button_release_event().connect (mem_fun (*this,&StepUI::value_display_button_release), false);
	valueDisplay.signal_key_press_event().connect (mem_fun (*this,&StepUI::value_display_key_press), false);
	valueDisplay.signal_changed().connect (mem_fun (*this,&StepUI::value_display_changed));

	valueDisplayFrame.set_shadow_type (Gtk::SHADOW_IN);
	valueDisplayFrame.add (valueDisplay);

	pack_start (valueDisplayFrame, false, false);

	terminalButton.set_size_request (10, 10);
	terminalButton.set_name ("TerminalMarker");
	terminalButton.show_all ();

	/* do things when we click on the buttons */

	terminalButton.signal_button_press_event().connect
		(mem_fun (*this,&StepUI::terminal_click));

	terminalButton.signal_expose_event().connect
		(mem_fun (*this, &StepUI::draw_terminal));
	
	/* do things when the adjustment is changed */

	adjustment.signal_value_changed().connect (mem_fun (*this, &StepUI::adjustment_changed));

	/* queue a touch/update whenever the step is changed */

	step.value_change.connect (mem_fun (*this, &StepUI::queue_touch));
	step.current_change.connect (mem_fun (*this, &StepUI::queue_touch));
	step.terminal_change.connect (mem_fun (*this, &StepUI::queue_touch));
	step.active_change.connect (mem_fun (*this, &StepUI::queue_touch));

	/* this will set and then unset `touched' so that we begin
	   in the correct state.
	*/

	map_state ();
}

void
StepUI::silently_set_adjustment (int val)

{
	GtkAdjustment *adj = adjustment.gobj();

	adj->lower = step.lower();
	adj->upper = step.upper();
	adj->value = val;
}

void
StepUI::adjustment_changed ()
{
	if (step.sequence().is_raw_midi()) {
		return;
	}

	if (adjustment.get_value() != (float) step.val()) {
		step.set ((int) adjustment.get_value());
		set_value_display ();
	}
}

void
StepUI::touch ()
{
	int val;
	bool cur;
	bool term;
	bool act;
	bool act_cur = false;


	if (!touched) {
		return;
	}

	touched = false;

	if (adjustment.get_value() != (val = step.val())) {
		set_value_display ();
		silently_set_adjustment (val);
	} 

	if (ui_is_current != (cur = step.is_current())) {
		act_cur = true;
	}

	if (ui_is_enabled != (act = step.is_enabled())) {
		act_cur = true;
	}

	if (act_cur) {
		set_current_active_display ();
		ui_is_enabled = act;
		ui_is_current = cur;
	}

	if (ui_is_terminal != (term = step.is_terminal())) {
		set_terminal_display ();
		ui_is_terminal = step.is_terminal();
	}
}

void
StepUI::queue_touch ()
{
	touched = true;
}

void
StepUI::map_state ()

{
	/* force updates on all displays */

	silently_set_adjustment (step.val() + 1);
	ui_is_current = !step.is_current();
	ui_is_terminal = !step.is_terminal ();
	ui_is_enabled = !step.is_enabled ();
	queue_touch ();

	if (step.sequence().is_raw_midi()) {
		SequencerUI::instance()->set_tip (&valueDisplay, _("Middle-click to edit the MIDI data for this step"));
	} else {
		SequencerUI::instance()->set_tip (&valueDisplay, "");
	}

	touch ();
}

/* ----- Button Press/Release Callbacks ------------------------- */

bool
StepUI::value_display_scroll_event (GdkEventScroll* ev)
{
	switch (ev->direction) {
	case GDK_SCROLL_UP:
		adjustment.set_value (adjustment.get_value() + adjustment.get_step_increment());
		break;
	case GDK_SCROLL_DOWN:
		adjustment.set_value (adjustment.get_value() - adjustment.get_step_increment());
		break;
	default:
		break;
	}
	return true;
}

bool
StepUI::value_display_button_press (GdkEventButton *ev)
{
	switch (ev->button) {
	case 3:
		/* we own button 3 */
		return true;
		
	case 2:
		/* we own button 2 */
		return true;
		
	}
	return false;
}

bool
StepUI::value_display_button_release (GdkEventButton *ev)
{
	switch (ev->button) {
	case 3:
		if (step.is_enabled()) {
			step.disable ();
		} else {
			step.enable ();
		}
		return true;
		break;

	case 2:
		if (step.sequence().is_raw_midi()) {
			MIDI::byte *msg;
			size_t len;
			char why[64];
			
			sprintf (why, "Sequence %d Step %d",
				 step.sequence().id() + 1,
				 step.id() + 1);

			string anal_cpp_compiler(why);

			msg = step.midimsg (&len);
			msg = SequencerUI::midibyte_editor()->edit (anal_cpp_compiler, &len, msg);
			step.set (msg, len);
			
		} else if (ev->state & GDK_CONTROL_MASK && ev->button == 2) {
			step.learn_about_external_control ();
		}
		return true;
		break;
	}

	return false;
}

gint
StepUI::terminal_click (GdkEventButton *ev)

{
	switch (ev->button) {
	case 1:
		step.sequence().set_first_step (step.id());
		break;

	case 2:
		if (step.is_end()) {
			step.sequence().set_last_step (step.id() + 1);
				 
		} else {
			step.sequence().set_first_step (step.id() - 1);
		}
		break;
			
	case 3:
		step.sequence().set_last_step (step.id());
		break;
	}

	return FALSE;
}

void
StepUI::set_terminal_display ()
{
	terminalButton.queue_draw ();
}

bool
StepUI::draw_terminal (GdkEventExpose* ev)
{
	int height;
	int width;
	int foo, bar, baz;

	terminalButton.get_window()->get_geometry (foo, bar, width, height, baz);

	cairo_t* cr = gdk_cairo_create(GDK_DRAWABLE(terminalButton.get_window()->gobj()));

	if (step.is_start()) {
		cairo_set_source_rgba (cr, 0.0, 0.0, 1.0, 0.8);
		cairo_move_to (cr, 0, 0);
		cairo_line_to (cr, width, height/2);
		cairo_line_to (cr, 0, height);
		cairo_line_to (cr, 0, 0);
		cairo_fill (cr);
	} else if (step.is_end ()) {
		cairo_set_source_rgba (cr, 0.0, 0.0, 1.0, 0.8);
		cairo_move_to (cr, width, 0);
		cairo_line_to (cr, width, height);
		cairo_line_to (cr, 0, height/2);
		cairo_line_to (cr, width, 0);
		cairo_fill (cr);
	} else {
		cairo_set_source_rgba (cr, 1.0, 0.0, 0.0, 0.2);
		cairo_move_to (cr, height/2, height/2);
		cairo_arc (cr, height/2, height/2, height/2, 0, 2.0 * M_PI);
		cairo_fill (cr);
	}

	cairo_destroy (cr);
	return true;
}

const char*
midi_note_name (size_t val)
{
	if (val > 127) {
		return "???";
	}

	static const char* notes[] = {
		"c",
		"c#",
		"d",
		"d#",
		"e",
		"f",
		"f#",
		"g",
		"a",
		"a#",
		"b",
		"b#"
	};

	int octave = val/12;
	static char buf[8];
	
	val -= octave*12;

	snprintf (buf, sizeof (buf), "%s%d", notes[val], octave);
	return buf;
}
	
void
StepUI::set_value_display ()

{
	if (step.sequence().is_raw_midi()) {

		if (step.midimsg()) {
			valueDisplay.set_text ("M");
		} else {
			valueDisplay.set_text ("?");
		}

	} else {
		char buf[16];
		
		if (step.sequence().makes_sound()) {
			sprintf (buf, "%s", midi_note_name (step.val()));
		} else {
			sprintf (buf, "%d", step.val());
		}
		valueDisplay.set_text (buf);
	} 
}

void
StepUI::set_current_active_display ()

{
	Gtk::StateType state;

	if (step.is_current()) {
		if (step.is_enabled()) {
			state = Gtk::STATE_ACTIVE;
		} else {
			state = Gtk::STATE_PRELIGHT;
		}
	} else {
		if (step.is_enabled()) {
			state = Gtk::STATE_NORMAL;
		} else {
			state = Gtk::STATE_PRELIGHT;
		}
	}	

	valueDisplay.set_state (state);
}

void
StepUI::learning_started ()
{
	learning_popup.set_text ("Operate MIDI controllers now");
	learning_popup.touch ();
}

void
StepUI::learning_stopped ()
{
	learning_popup.touch ();
}

gint
StepUI::pointer_entered_step (GdkEventCrossing *ev)
{
	button_release_can_toggle = true;
	return FALSE;
}


gint
StepUI::pointer_left_step (GdkEventCrossing *ev)
{
	button_release_can_toggle = false;
	return FALSE;
}
 
bool
StepUI::value_display_key_press (GdkEventKey *ev) 
{
	if (step.sequence().is_raw_midi()) {
		/* no key entry */
		return true;
	}

	switch (ev->keyval) {
	case GDK_Up:
	case GDK_period:
		if (ev->state & GDK_SHIFT_MASK) {
			step.set (step.val() + 12);
		} else {
			step.set (step.val() + 1);
		}
		return true;
		break;

	case GDK_numbersign:
	case GDK_musicalsharp:
		step.set (step.val() + 1);
		return true;
		
	case GDK_Down:
		if (ev->state & GDK_SHIFT_MASK) {
			step.set (step.val() - 12);
		} else {
			step.set (step.val() - 1);
		}
		return true;
		break;

	case GDK_comma:
	case GDK_musicalflat:
		step.set (step.val() - 1);
		return true;

	case GDK_c:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 0);
			return true;
		}
		return false;

	case GDK_d:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 2);
			return true;
		}
		return false;

	case GDK_e:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 4);
			return true;
		}
		return false;

	case GDK_f:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 5);
			return true;
		}
		return false;

	case GDK_g:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 7);
			return true;
		}
		return false;

	case GDK_a:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 9);
			return true;
		}
		return false;

	case GDK_b:
		if (step.sequence().makes_sound()) {
			step.set (step.val()/12 * 12 + 11);
			return true;
		}
		return false;

	case GDK_0:
	case GDK_1:
	case GDK_2:
	case GDK_3:
	case GDK_4:
	case GDK_5:
	case GDK_6:
	case GDK_7:
	case GDK_8:
	case GDK_9:
		return false;

	case GDK_Left:
	case GDK_Right:
	case GDK_Tab:
		return false;

	default:
		/* dump any other chars */
		return true;
	}
}

void
StepUI::value_display_changed ()
{
	float value;

	if (step.sequence().is_raw_midi()) {
		return;
	}

	string str = valueDisplay.get_text ();

	if (str.length() == 0) {
		return;
	}

	if (sscanf (str.c_str(), "%f", &value) == 0) {
		return;
	}

	adjustment.set_value (value);
}

gint
StepUI::focus_in_step(GdkEventFocus* ev)
{
	valueDisplay.select_region (0, -1);
	return TRUE;
}

gint
StepUI::focus_out_step(GdkEventFocus* ev)
{
	valueDisplay.select_region (0, 0);
	return FALSE;
}
