/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: trigger.cc,v 1.3 2000/05/03 15:54:21 pbd Exp $
*/

#include <iostream>

#include <midi++/port.h>
#include <midi++/manager.h>
#include <pbd/error.h>

#include "trigger.h"

using namespace MIDI;
using namespace PBD;

const char *Trigger::type_strings[] = {
	"Note On",
	"Note Off",
	"Controller",
	"Raw MIDI",
	0
};


const char *Trigger::source_strings[] = {
	"Input",
	"Output",
	0
};

Trigger::Trigger ()

{
	/* default settings for a new trigger. Note that the trigger
	   is not connected until one of these values is changed or
	   reset() is called.
	*/

	_port = MIDI::Manager::instance()->inputPort;
	_channel = 0;
	_type = note_on;
	_source = input;
	_value = 60;
	_controller_id = 1;

	_msg = 0;
	current_connection = -1;
	current_source = -1;
	length = 0;
	matched_so_far = 0;
}

Trigger::~Trigger ()

{
	deleted (*this);

	connection.disconnect ();

	if (_msg) {
		delete [] _msg;
	}
}

void
Trigger::reset ()
{
	matched_so_far = 0;
	connect ();
}

void
Trigger::set_name (const std::string &str)

{
	std::string oldname = _name;
	_name = str;

	name_changed (*this, oldname);
}

void
Trigger::set_type (Type t)

{
	_type = t;
	connect ();
	changed (*this);
}

void
Trigger::set_source (Source s)

{
	_source = s;
	connect ();
	changed (*this);
}

void
Trigger::set_value (MIDI::byte val)

{
	_value = val;
	connect ();
	changed (*this);
}

void
Trigger::set_controller_id (MIDI::byte cid)

{
	_controller_id = cid;
	connect ();
	changed (*this);
}

void
Trigger::set_midibytes (MIDI::byte *msg, size_t len)

{
	if (msg != _msg && _msg) {
		delete [] msg;
	}

	_msg = msg;
	length = len;
	reset ();
	changed (*this);
}

void
Trigger::set_port (MIDI::Port *p)

{
	_port = p;
	reset ();
	changed (*this);
}

void
Trigger::set_channel (MIDI::channel_t chn)

{
	_channel = chn;
	reset ();
	changed (*this);
}

void
Trigger::connect ()
{
	if (_port == 0) {
		return;
	}

	if (current_connection == (int) _type &&
	    current_source == (int) _source) {
		return;
	}

	connection.disconnect ();

	switch (_type) {
	case note_on:
		connect_note_on ();
		break;
	case note_off:
		connect_note_off ();
		break;
	case controller:
		connect_controller ();
		break;
	case midibytes:
		if (_msg != 0 && length != 0) {
			connect_midibytes ();
		} else {
			/* defer till we know the message */
			return;
		}
	}

	current_connection = _type;
	current_source = _source;
}
		
void
Trigger::connect_note_on () 
	
{
	Parser *p;

	if (_source == input) {
		p = _port->input();
	} else {
		p = _port->output();
	}

	connection = p->channel_note_on[(int)_channel].connect 
		(mem_fun (*this, &Trigger::handle_note));
}

void
Trigger::connect_note_off ()

{
	Parser *p;

	if (_source == input) {
		p = _port->input();
	} else {
		p = _port->output();
	}

	connection = p->channel_note_off[(int)_channel].connect 
			(mem_fun (*this, &Trigger::handle_note));
}

void
Trigger::connect_controller ()

{
	Parser *p;

	if (_source == input) {
		p = _port->input();
	} else {
		p = _port->output();
	}

	connection = p->channel_controller[(int)_channel].connect (mem_fun (*this, &Trigger::handle_controller));
}

void
Trigger::connect_midibytes ()

{
	Parser *p;

	if (_source == input) {
		p = _port->input();
	} else {
		p = _port->output();
	}

	connection = p->raw_preparse.connect (mem_fun (*this, &Trigger::handle_midibyte));
}

void
Trigger::handle_note (Parser &p, MIDI::EventTwoBytes *ev)

{
	if (ev->note_number == _value) {
		fired (*this);
	}
}

void
Trigger::handle_controller (Parser &p, MIDI::EventTwoBytes *ev)

{
	if (ev->controller_number == _controller_id && ev->value == _value) {
		fired (*this);
	}
}

void
Trigger::handle_midibyte (Parser &p, byte *msg, size_t msglen)

{
	if (_msg && _msg[matched_so_far] == *msg) {
		matched_so_far++;
		if (matched_so_far == length) {
			matched_so_far = 0;
			fired (*this);
		}
	} else {
		matched_so_far = 0;
	}
}

void
Trigger::dump (std::ostream &o)

{
	unsigned int i;
    
	o << '"' << _name << '"' << ' '
	  << (unsigned int) _type << ' '
	  << (unsigned int) _source << ' '
	  << '"' << _port->name () << '"' << ' '
	  << length << ' '
	  << (int) _value << ' '
	  << (int) _controller_id << ' '
	  << (int) _channel << ' ';
	
	for (i = 0; i < length; i++) {
		o << (int) _msg[i] << ' ';
	}
}     

void
Trigger::load (std::istream &o)

{
	int i;
	char c;
	char buf[64];
	std::string portname;

	if (_msg) {
		delete [] _msg;
		length = 0;
		_msg = 0;
	}
	
	o >> c; /* eat " */
	o.getline (buf, sizeof (buf), '"');
	_name = buf;
	o >> i; _type = (Type) i;
	o >> i; _source = (Source) i;
	o >> c; /* eat " */
	o.getline (buf, sizeof (buf), '"');
	portname = buf;
	o >> length;
	o >> i; _value = (byte) i;
	o >> i; _controller_id = (byte) i;
	o >> i; _channel = (channel_t) i; 

	if (length) {
		_msg = new byte [length];
	
		for (size_t n = 0; n < length; n++) {
			o >> i; _msg[n] = (byte) i;
		}
	}

	if ((_port = MIDI::Manager::instance()->port (portname)) == 0){
		error << "Port \""
		      << portname
		      << "\" use for trigger is unknown"
		      << endmsg;
		return;
	}

	switch (_type) {
	case note_on:
		connect_note_on ();
		break;
	case note_off:
		connect_note_off ();
		break;
	case controller:
		connect_controller ();
		break;
	case midibytes:
		connect_midibytes ();
		break;
	}
}     

std::ostream& 
operator<< (std::ostream &ostr, Trigger &t) 
{
	t.dump (ostr);
	return ostr;
}

std::istream& 
operator>> (std::istream &istr, Trigger &t)
{
	t.load (istr);
	return istr;
}

Trigger::Type
Trigger::type_string_to_type (const std::string &str)

{
	int i;

	for (i = 0; type_strings[i]; i++) {
		if (str == type_strings[i]) {
			return (Type) i;
		}
	}

	fatal << "unknown type string \""
	      << str
	      << '"'
	      << endmsg;

	/*NOTREACHED*/
	return note_on; /* keep compiler happy */
}

Trigger::Source
Trigger::source_string_to_source (const std::string &str)

{
	int i;

	for (i = 0; source_strings[i]; i++) {
		if (str == source_strings[i]) {
			return (Source) i;
		}
	}

	fatal << "unknown source string \""
	      << str
	      << '"'
	      << endmsg;

	/*NOTREACHED*/
	return input; /* keep compiler happy */
}

const char *
Trigger::type_string ()

{
	return type_strings[(int)_type];
}

const char *
Trigger::source_string ()

{
	return source_strings[(int)_source];
}
