/*
    Copyright (C) 1998-2007 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: trigger.h,v 1.3 2000/05/03 15:54:21 pbd Exp $
*/

#ifndef __sw_trigger_h__
#define __sw_trigger_h__

#include <string>
#include <iostream>
#include <sigc++/signal.h>
#include <sigc++/bind.h>
#include <midi++/types.h>

class Sequencer;

namespace MIDI {
	class Port;
	class Parser;
}

class Trigger : public sigc::trackable

{
  public:
	enum Type { 
		note_on = 0,
		note_off = 1,
		controller = 2,
		midibytes = 3
	};

	enum Source { 
		input = 0,
		output = 1
		/* XXX maybe add "in_and_output" and "time" later */
	};

	static const char *type_strings[];
	static const char *source_strings[];

	Trigger ();
	~Trigger ();

	void reset (); /* re-arm, potentially reconnect */

	static Type type_string_to_type (const std::string &);
	static Source source_string_to_source (const std::string &);

	void set_name (const std::string &str);
	void set_type (Type);
	void set_source (Source);
	void set_port (MIDI::Port *);
	void set_channel (MIDI::channel_t);
	void set_value (MIDI::byte);
	void set_controller_id (MIDI::byte);
	void set_midibytes (MIDI::byte *msg, size_t msglen);

	const std::string   &name ()           { return _name; }
	Type            type ()           { return _type; }
	Source          source ()         { return _source; }
	MIDI::Port     *port ()           { return _port; }
	MIDI::channel_t channel ()        { return _channel; }
	MIDI::byte     *msg (size_t *len) { *len = length; return _msg; }
	MIDI::byte      value()           { return _value; }
	MIDI::byte      controller_id()   { return _controller_id; }
	const char     *type_string ();
	const char     *source_string ();

	sigc::signal1<void,Trigger&> fired;
	sigc::signal1<void,Trigger&> deleted;
	sigc::signal1<void,Trigger&> changed;
	sigc::signal2<void,Trigger&,const std::string&> name_changed;

  protected:
	void dump (std::ostream &);
	void load (std::istream &);
	friend std::ostream& operator<< (std::ostream &, Trigger &);
	friend std::istream& operator>> (std::istream &, Trigger &);

  private:
	MIDI::Port *_port;
	MIDI::channel_t _channel;
	MIDI::byte _value;
	MIDI::byte _controller_id;
	MIDI::byte *_msg;
	std::string _name;
	Source _source;
	Type _type;
	unsigned int length;
	unsigned int matched_so_far;
	int current_connection; /* almost like a Type, but can be -1 */
	int current_source; /* almost like a Source, but can be -1 */

	sigc::connection connection;
	
	void connect ();
	
	void handle_midibyte (MIDI::Parser &, MIDI::byte *buf, size_t bufsize);
	void handle_note (MIDI::Parser &, MIDI::EventTwoBytes *);
	void handle_controller (MIDI::Parser &, MIDI::EventTwoBytes *);

	void connect_note_on ();
	void connect_note_off ();
	void connect_controller ();
	void connect_midibytes ();
};

std::ostream& operator<< (std::ostream &, Trigger &);
std::istream& operator>> (std::istream &, Trigger &);

#endif /* __sw_trigger_h__ */
