/*
    Copyright (C) 1999-2007 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: trigger_ui.cc,v 1.2 2000/05/02 02:47:16 pbd Exp $
*/

#include <pbd/error.h>
#include <midi++/manager.h>
#include <midi++/types.h>
#include <gtkmm2ext/utils.h>

#include "trigger.h"
#include "trigger_ui.h"
#include "ui.h"

using namespace PBD;
using namespace Gtk;

bool TriggerUI::have_descriptive_strings = false;
vector<string> TriggerUI::source_strings;
vector<string> TriggerUI::type_strings;
static const char *longest_source_string = "";
static const char *longest_type_string = "";

TriggerUI::TriggerUI (Trigger &trig, bool with_delete)
	: _trigger (trig),
	  channelAdjustment (1, 1, 16, 1, 0),
	  channelSelector (channelAdjustment),
	  valueAdjustment (0, 0, 127, 1, 16),
	  valueSelector (valueAdjustment)
{
	if (!have_descriptive_strings) {
		get_descriptive_strings ();
	}

	set_title ("SoftWerk: Trigger Editor");

	get_vbox()->pack_start (nameBox, false, false, 10);
	get_vbox()->pack_start (typeBox, false, false);
	get_vbox()->pack_start (portBox, false, false);
	get_vbox()->pack_start (numbersBox, false, false);
	
	nameBox.pack_start (nameEntryBox, true, true, 25);

	typeBox.pack_start (typeSelectorBox, true, false);
	typeBox.pack_start (sourceSelectorBox, true, false);

	portBox.pack_start (portSelectorBox, true, false);
	portBox.pack_start (channelSelectorBox, true, false);
 
	numbersBox.pack_start (controllerSelectorBox, true, false);
	numbersBox.pack_start (valueSelectorBox, true, false);

	add_button (Stock::OK, RESPONSE_ACCEPT);
	if (with_delete) {
		add_button (Stock::DELETE, RESPONSE_REJECT);
	}
	add_button (Stock::CANCEL, RESPONSE_CANCEL);

	valueSelectorBox.pack_start (valueSelectorLabel, false, false);
	valueSelectorBox.pack_start (valueSelector, false, false);

	controllerSelectorBox.pack_start (controllerSelectorLabel,false,false);
	controllerSelectorBox.pack_start (controllerSelector, false, false);
	
	portSelectorBox.pack_start (portSelectorLabel, false, false);
	portSelectorBox.pack_start (portSelector, false, false);

	channelSelectorBox.pack_start (channelSelectorLabel, false, false);
	channelSelectorBox.pack_start (channelSelector, false, false);
	
	typeSelectorBox.pack_start (typeSelectorLabel, false, false);
	typeSelectorBox.pack_start (typeSelector, false, false);

	sourceSelectorBox.pack_start (sourceSelectorLabel, false, false);
	sourceSelectorBox.pack_start (sourceSelector, false, false);

	nameEntryBox.pack_start (nameEntryLabel, false, false);
	nameEntryBox.pack_start (nameEntry, false, false);

	nameEntryLabel.set_text ("Trigger Name");
	sourceSelectorLabel.set_text ("Event Source");
	typeSelectorLabel.set_text ("Event Type");
	channelSelectorLabel.set_text ("Channel");
	portSelectorLabel.set_text ("Event Port");
	controllerSelectorLabel.set_text ("Controller ID");

	/* combo selectors need strings */

	Gtkmm2ext::set_popdown_strings (typeSelector, type_strings);
	Gtkmm2ext::set_popdown_strings (sourceSelector, source_strings);
	Gtkmm2ext::set_popdown_strings (portSelector, SequencerUI::port_strings());
	Gtkmm2ext::set_popdown_strings (controllerSelector, SequencerUI::controller_strings ());

	/* spinners are all integers */

	show_all ();

	/* project current state into UI */

	map ();

	/* Hook up to relevant signals */

	typeSelector.signal_changed().connect (mem_fun (*this,&TriggerUI::type_change));
	sourceSelector.signal_changed().connect (mem_fun (*this,&TriggerUI::source_change));
	portSelector.signal_changed().connect (mem_fun (*this,&TriggerUI::port_change));
	channelAdjustment.signal_value_changed().connect (mem_fun (*this,&TriggerUI::channel_change));
	valueAdjustment.signal_value_changed().connect(mem_fun (*this,&TriggerUI::value_change));
	controllerSelector.signal_changed().connect (mem_fun (*this,&TriggerUI::controller_change));
	nameEntry.signal_changed().connect(mem_fun (*this,&TriggerUI::name_change));
}

void
TriggerUI::map ()

{
	nameEntry.set_text (_trigger.name());

	switch (_trigger.type()) {
	case Trigger::note_on:
	case Trigger::note_off:
		controllerSelectorBox.hide();
		valueSelectorBox.show ();
		valueSelectorLabel.set_label ("Note");
		channelSelectorBox.show ();
		break;

	case Trigger::controller:
		controllerSelectorBox.show ();
		valueSelectorBox.show ();
		valueSelectorLabel.set_label ("Value");
		channelSelectorBox.show ();
		break;
		
	case Trigger::midibytes:
		controllerSelectorBox.hide ();
		valueSelectorBox.hide ();
		channelSelectorBox.hide ();
		break;
	}

	typeSelector.set_active_text (_trigger.type_string());
	sourceSelector.set_active_text (_trigger.source_string());

	if (_trigger.type () != Trigger::midibytes) {
		valueSelector.get_adjustment()->set_value (_trigger.value ());
		channelSelector.get_adjustment()->set_value (_trigger.channel());
		portSelector.set_active_text (_trigger.port()->name ());

		if (_trigger.type() == Trigger::controller) {
			controllerSelector.set_active_text (MIDI::controller_names[_trigger.controller_id ()]);
		}
	}
}

void
TriggerUI::name_change ()
{	
	string new_name = nameEntry.get_text ();

	if (!new_name.empty()) {
		_trigger.set_name (new_name);
	}
}

void
TriggerUI::port_change ()
{
	MIDI::Port *port;
	string new_port;

	new_port = portSelector.get_active_text ();
	if ((port = MIDI::Manager::instance()->port (new_port)) == 0) {
		fatal << "unknown but named port \"" 
		      << new_port
		      << '"'
		      << endmsg;
	}

	_trigger.set_port (port);
}

void
TriggerUI::channel_change ()
{
	MIDI::channel_t new_channel;

	new_channel = (MIDI::channel_t) 
		channelSelector.get_adjustment()->get_value ();

	/* do the usual MIDI channel/UI correction */

	new_channel -= 1;

	_trigger.set_channel (new_channel);
}

void
TriggerUI::type_change ()
{
	Trigger::Type t;
	string new_type = typeSelector.get_active_text ();

	t = Trigger::type_string_to_type (new_type);
	_trigger.set_type (t);

	map ();
}

void
TriggerUI::source_change ()
{
	Trigger::Source s;
	string new_source = sourceSelector.get_active_text ();

	s = Trigger::source_string_to_source (new_source);
	_trigger.set_source (s);
}

void 
TriggerUI::value_change ()
{
	MIDI::byte newval = (MIDI::byte)
		valueSelector.get_adjustment()->get_value ();

	_trigger.set_value (newval);
}

void
TriggerUI::controller_change ()
{
	string name = controllerSelector.get_active_text();
	MIDI::byte newid = MIDI::decode_controller_name (name.c_str());
	_trigger.set_controller_id (newid);
}

void
TriggerUI::get_descriptive_strings ()

{
	int n;

	for (n = 0; Trigger::type_strings[n]; n++) {
		const char *str;

		str = Trigger::type_strings[n];
		
		if (strlen (str) > strlen (longest_type_string)) {
			longest_type_string = str;
		}
		type_strings.push_back ((char *) str);
	}

	for (n = 0; Trigger::source_strings[n]; n++) {
		const char *str;

		str = Trigger::source_strings[n];
		
		if (strlen (str) > strlen (longest_source_string)) {
			longest_source_string = str;
		}
		source_strings.push_back ((char *) str);
	}

	have_descriptive_strings = true;
}
