/*
    Copyright (C) 1998-2007 Paul Davis 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: ui.cc,v 1.11 2000/11/25 00:10:31 pbd Exp $
*/

#include <algorithm>

#include <math.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <iostream>

#include <gtkmm.h>
#include <gtk/gtkaccelmap.h>
#include <gtkmm/accelmap.h>
#include <gtkmm/stock.h>
#include <pbd/stl_delete.h>
#include <pbd/error.h>
#include <pbd/failed_constructor.h>
#include <midi++/port.h>
#include <midi++/manager.h>
#include <gtkmm2ext/textviewer.h>
#include <gtkmm2ext/pix.h>
#include <gtkmm2ext/utils.h>

#include "ui.h"
#include "sequence.h"
#include "sequencer.h"
#include "step.h"
#include "actions.h"
#include "patternfile_ui.h"
#include "midibyte_edit.h"
#include "configuration.h"
#include "trigger_ui.h"

#include "i18n.h"

using namespace Gtk;
using namespace PBD;

SequencerUI *SequencerUI::theSequencerUI = 0;
MidiByteEditor *SequencerUI::theMidiEditor = 0;


vector<string> SequencerUI::_port_strings;
vector<string> SequencerUI::_controller_strings;
vector<string> SequencerUI::clock_mode_strings;
const char *SequencerUI::longest_port_string = "";
const char *SequencerUI::longest_controller_string = "";

Glib::RefPtr<Gdk::Pixbuf> SequencerUI::knob_pixbuf;

SequencerUI::SequencerUI (int *argc, char ***argv, const char *rc_file)
	: Gtkmm2ext::UI ("SoftWerk", argc, argv),
	  _sequencer (0),
	  /* XXX use Config->default_bpm */
	  clock_adjustment (120, 1, 300, 1, 10),
	  clock_spinner (&clock_adjustment, "ClockDisplay"),
	  tempocontrol (clock_adjustment),
	  transport_button ("gtk-media-play"),
	  rew_button ("gtk-media-rewind")

{
	if (theSequencerUI == 0) {
		theSequencerUI = this;
		theMidiEditor = new MidiByteEditor;
	}

	get_descriptive_strings ();

	load_rcfile (rc_file, false);
}

gint
SequencerUI::close (GdkEventAny *ev)
{
	quit ();
	return true;
}

static void 
accel_map_changed (GtkAccelMap* map,
		   gchar* path,
		   guint  key,
		   GdkModifierType mod,
		   gpointer arg)
{
	static_cast<SequencerUI*>(arg)->save_keybindings ();
}

int	
SequencerUI::setup_windows ()
{
	if (setup_pixmaps ()) {
		return -1;
	}

	ActionManager::init ();
	add_actions ();

	main_window.add (vpacker);
	main_window.set_size_request (1010, 750);

	main_window.signal_delete_event().connect (mem_fun (*this,&SequencerUI::close));
	
	vpacker.set_homogeneous (false);
	sequence_vpacker.set_homogeneous (false);

	menubar = make_menubar ();

	/* Set up clock spinner and mode select button */

	clock_mode = internal;
	clock_spinner.set_print_func (clock_filter, this);
	clock_spinner.get_adjustment().signal_value_changed().connect (mem_fun (*this,&SequencerUI::clock_changed));
	clock_is_internal ();
	clock_frame.set_shadow_type (Gtk::SHADOW_IN);
	clock_frame.add (clock_spinner);
	clockBox.pack_start (clock_frame);
	Gtkmm2ext::set_size_request_to_display_given_text (clock_spinner, "10000 BPM: 1000 ticks/beat", 10, 10);

	transport_button.set_use_stock (true);
	rew_button.set_use_stock (true);

	Gtk::HBox* transportcontrols = manage (new HBox());
	transportcontrols->pack_start (transport_button);
	transportcontrols->pack_start (rew_button);

	ActionManager::get_action ("transport", "toggle")->connect_proxy (transport_button);
	ActionManager::get_action ("transport", "rew")->connect_proxy (rew_button);

	Gtk::HBox* clock_mode_box = manage (new HBox());
	Gtk::Label* clock_mode_label = manage (new Label (_("Clock mode")));

	Gtkmm2ext::set_popdown_strings (clock_selector, clock_mode_strings);
	clock_selector.set_active_text (clock_mode_strings.front());
	clock_selector.signal_changed().connect (mem_fun (*this, &SequencerUI::clock_mode_changed));

	clock_mode_box->pack_start (*clock_mode_label, false, false, 10);
	clock_mode_box->pack_start (clock_selector);

	tempocontrol.set_draw_value (false);
	tempocontrol.set_size_request (150,-1);

	top_row.set_spacing (5);

	top_row.pack_start (*menubar, false, false);

	VBox* tempocontrolbox = manage (new VBox());
	Label* tempocontrollabel = manage (new Label (_("Tempo")));

	tempocontrolbox->pack_start (*tempocontrollabel);
	tempocontrolbox->pack_start (tempocontrol);

	top_row.pack_end (clockBox, false, false, 10);
	top_row.pack_end (*tempocontrolbox, false, false, 10);
	top_row.pack_end (*clock_mode_box, false, false, 10);
	top_row.pack_end (*transportcontrols, false, false, 10);

	vpacker.pack_start (top_row, false, false, 5);

	scroller_vpacker.pack_start (sequence_scroller);
	sequence_scroller.add (sequence_vpacker);
	sequence_scroller.set_policy (Gtk::POLICY_NEVER, Gtk::POLICY_AUTOMATIC);

	vpacker.pack_start (scroller_vpacker);

	main_window.show_all ();

	/* add each sequenceUI to the display */

	_sequencer->foreach_sequence (mem_fun (*this, &SequencerUI::add_sequence));

	/* hook ourselves up to some "global" signals */

	_sequencer->tick.connect (mem_fun (*this, &SequencerUI::send_tick));
	_sequencer->reloaded.connect (mem_fun (*this, &SequencerUI::map_state));
	_sequencer->sequence_added.connect (mem_fun (*this, &SequencerUI::add_sequence));

	/* now project the state of the sequencer onto the UI */

	map_state ();

	rebuild_trigger_submenu ();

	try {
		AccelMap::load (Config->keybindings_path);
	} catch (...) {
		error << string_compose (_("SoftWerk key bindings file not found at \"%1\" or contains errors."), Config->keybindings_path)
		      << endmsg;
	}

	/* catch changes */

	GtkAccelMap* accelmap = gtk_accel_map_get();
	g_signal_connect (accelmap, "changed", (GCallback) accel_map_changed, this);

	main_window.add_accel_group (ActionManager::ui_manager->get_accel_group());

	return 0;
}

void
SequencerUI::save_keybindings ()
{
	AccelMap::save (Config->keybindings_path);
}

void
SequencerUI::map_state ()

{
	vector<SequenceUI *>::iterator i;

	/* The clock is the only global thing we need to map into the UI 
	*/

	/* The clock value is in BPM. Convert from internal usecs_per_tick.
	*/

	double bpm;
	
	bpm = (60.0 * 1000000.0) / (_sequencer->current_usecs_per_tick () * Config->ticks_per_beat);

	clock_spinner.get_adjustment().set_value (bpm);
	set_clock_mode (internal);

	for (i = sequences.begin(); i != sequences.end(); i++) {
		(*i)->map_state ();
	}
}

int
SequencerUI::setup_pixmaps ()
{
	try {

		knob_pixbuf = Gdk::Pixbuf::create_from_file (Config->knob_image);

	}  catch (...) {
		
		error << "No knob image found (or not loadable) at "
		      << Config->knob_image
		      << endmsg;
		return -1;
	}

	return 0;
}

void
SequencerUI::_add_sequence (Sequence &s, void *arg)
{
	((SequencerUI *) arg)->add_sequence (s);
}

void
SequencerUI::add_sequence (Sequence &s)
{
	SequenceUI *sui;
	vector<SequenceUI *>::iterator i;

	sui = new SequenceUI (s);
	sequence_vpacker.pack_start (*sui, false, false);
	sequences.push_back (sui);

	/* tell this and all pre-existing sequence UI's to 
	   add a track control button for this sequence.
	*/

	for (i = sequences.begin(); i != sequences.end (); i++) {
		(*i)->add_track_button (s.id());
	}

}

int
SequencerUI::setup_tick_pipe ()
{
	if (pipe (tick_pipe)) {
		error << "UI: cannot create tick pipe ("
		      << strerror (errno) << ")" 
		      << endmsg;

		return -1;
	}

	if (fcntl (tick_pipe[0], F_SETFL, O_NONBLOCK)) {
		error << "UI: cannot set O_NONBLOCK on tick read pipe ("
		      << strerror (errno) << ")"
		      << endmsg;
		return -1;
	}

	if (fcntl (tick_pipe[1], F_SETFL, O_NONBLOCK)) {
		error << "UI: cannot set O_NONBLOCK on tick write pipe ("
		      << strerror (errno) 
		      << ")" 
		      << endmsg;
		return -1;
	}

	/* add the pipe to the select/poll loop that GDK does */

	gdk_input_add (tick_pipe[0],
		       GDK_INPUT_READ,
		       SequencerUI::tick_pipe_callback,
		       this);

	return 0;
}

void
SequencerUI::send_tick (unsigned int current_tick)
{
	char c;

	/* we don't care about sending the current tick. we also
	   don't care if the write fails. Just try it, and get out.
	*/
	write (SequencerUI::instance()->tick_pipe[1], &c, 1);
}

void
SequencerUI::tick_pipe_callback (void *arg, int fd, GdkInputCondition cond)
{
	char buf[64];
	int nread;

	/* empty the pipe */

	while (1) {
		if ((nread = read (((SequencerUI *) arg)->tick_pipe[0], 
				   buf, sizeof(buf))) < 0) {

			if (errno != EAGAIN) {
				error << "SequencerUI: "
					 "read error on tick pipe ("
				      << strerror (errno)
				      << endmsg;
				return;
			}

		} else if (nread == 0 || nread < (int) sizeof (buf)) {
			
			/* the pipe is now empty */

			break;
		}
	}

	((SequencerUI *) arg)->do_tick ();
}

void
SequencerUI::do_tick ()
{
	vector<SequenceUI *>::iterator s;

	for (s = sequences.begin(); s != sequences.end(); s++) {
		(*s)->touch ();
	}
}

void
SequencerUI::play()
{
	_sequencer->play ();
}

void
SequencerUI::stop()
{
	_sequencer->pause ();
}

void
SequencerUI::toggle_pause ()
{
	_sequencer->toggle_pause ();

	if (_sequencer->get_state() == Sequencer::running) {
		ActionManager::get_action ("transport", "toggle")->property_stock_id() = Stock::MEDIA_PAUSE;
	} else {
		ActionManager::get_action ("transport", "toggle")->property_stock_id() = Stock::MEDIA_PLAY;
	} 
}

void
SequencerUI::clock_mode_changed ()
{
	if (clock_selector.get_active_text() == clock_mode_strings.front()) {
		set_clock_mode (internal);
	} else {
		set_clock_mode (external);
	}
}

void
SequencerUI::clock_is_external ()
{
	clock_spinner.set_sensitive (false);
	tempocontrol.set_sensitive (false);
	ActionManager::set_sensitive (ActionManager::transport_actions, false);
}

void
SequencerUI::clock_is_internal ()
{
	clock_spinner.set_sensitive (true);
	tempocontrol.set_sensitive (true);
	ActionManager::set_sensitive (ActionManager::transport_actions, true);
}

void
SequencerUI::set_clock_mode (ClockMode mode)
{
	static bool first_time = true;

	if (mode == external) {
		clock_mode = external;
		clock_is_external ();
		_sequencer->use_MIDI_clock (true);
	} else {
		clock_mode = internal;
		clock_is_internal ();
		_sequencer->use_MIDI_clock (false);
	} 
	
	/* this will reprint the clock display without a value change */

	clock_spinner.set_print_func (clock_filter, this);

	/* now reset the sequencer clock */

	if (!first_time) {
		clock_changed ();
	} 

	first_time = false;
}

void
SequencerUI::clock_changed ()
{
	gfloat bpm;
	gfloat tpm;
	size_t upt;

	bpm = clock_spinner.get_adjustment().get_value ();

	/* The clock value is BPM. We have to convert to
	   usecs_per_tick, using the current ticks_per_beat
	   value.
	*/

	tpm = bpm * Config->ticks_per_beat;
	upt = (size_t) floor ((60.0 * 1000000.0) / tpm);

	switch (clock_mode) {
	case internal:
		_sequencer->set_usecs_per_tick (upt);
		break;

	default:
		break;
	}
}

void
SequencerUI::new_sequence ()
{
	Sequence *seq;
	
	seq = _sequencer->new_sequence ();

	if (seq) {
		add_sequence (*seq);
	}
}


Gtk::Menu*
SequencerUI::get_trigger_submenu ()
{
	Widget* foo = ActionManager::get_widget ("/Main/Triggers/Edit");
	Gtk::MenuItem* foomi = dynamic_cast<MenuItem*>(foo);
	return foomi->get_submenu ();
}

int
SequencerUI::remove_trigger_item (Trigger &trigger)
{
	using namespace Menu_Helpers;
	Gtk::Menu* trigger_menu = get_trigger_submenu();
	MenuList& trigger_list = trigger_menu->items();
	MenuList::iterator i;

	for (i = trigger_list.begin(); i != trigger_list.end(); i++) {
		if ((*i).get_data ("trigger_ptr") == &trigger) {
			trigger_list.erase (i);
			break;
		}
	}
		
	return 0;
}

int
SequencerUI::add_trigger_item (Trigger &trigger)
{
	using namespace Menu_Helpers;
	Gtk::Menu* trigger_menu = get_trigger_submenu();
	MenuList& trigger_list = trigger_menu->items();

	trigger_list.push_back (MenuElem (trigger.name().c_str(), bind (mem_fun (*this, &SequencerUI::edit_trigger), &trigger)));
	trigger_list.back().set_data ("trigger_ptr", &trigger);

	return 0;
}

void
SequencerUI::trigger_deleted (Trigger &trigger)
{
	remove_trigger_item (trigger);
	rebuild_trigger_submenu ();
}

void
SequencerUI::trigger_name_changed (Trigger &trigger, const string &oldname)
{
	remove_trigger_item (trigger);
	add_trigger_item (trigger);
}

void
SequencerUI::rebuild_trigger_submenu ()
{
	using namespace Menu_Helpers;
	Gtk::Menu* trigger_menu = get_trigger_submenu();
	MenuList& items = trigger_menu->items();

	items.clear ();

	_sequencer->foreach_trigger (*this, &SequencerUI::add_trigger_item);

	trigger_menu->show_all ();
}

Gtk::MenuBar*
SequencerUI::make_menubar ()
{
	return (Gtk::MenuBar*) ActionManager::get_widget ("/Main");
}

void
SequencerUI::add_trigger ()
{
	Trigger *trigger;

	char buf[32];

	trigger = new Trigger;

	sprintf (buf, "Trigger %zd", _sequencer->ntriggers()+1);
	trigger->set_name (buf);

	TriggerUI tui (*trigger, false); // no delete option

	switch (tui.run()) {
	case RESPONSE_ACCEPT:
		break;

	default:
		/* get rid of the newly created trigger */
		delete trigger;
		return;
	}

	trigger->reset ();

	_sequencer->add_trigger (*trigger);

	trigger->name_changed.connect (mem_fun (*this, &SequencerUI::trigger_name_changed));
	trigger->deleted.connect (mem_fun (*this, &SequencerUI::trigger_deleted));
	
	rebuild_trigger_submenu ();
}	
	
void
SequencerUI::edit_trigger (Trigger *trigger)
{
	TriggerUI tui (*trigger, true); // allow delete
	
	int response = tui.run ();

	switch (response) {
	case RESPONSE_ACCEPT:
		break;

	case RESPONSE_REJECT: // means delete
		delete trigger;
		return;

	default:
		return;
	}
	
	trigger->reset ();
}

void
SequencerUI::open_pattern_file ()
{
	FileChooserDialog fs (main_window, "Pattern File Selector", FILE_CHOOSER_ACTION_OPEN);
	
	FileFilter pattern_filter;
	pattern_filter.add_pattern ("*.swp");
	pattern_filter.set_name (_("SoftWerk patterns"));
	fs.add_filter (pattern_filter);
	fs.set_filter (pattern_filter);
	fs.set_current_folder (Config->pattern_directory);

	fs.add_button (Stock::OK, RESPONSE_ACCEPT);
	fs.add_button (Stock::CANCEL, RESPONSE_CANCEL);

	switch (fs.run()) {
	case RESPONSE_ACCEPT:
		break;
	default:
		return;
	}

	Config->pattern_directory = fs.get_current_folder ();
	
	Glib::ustring path = fs.get_filename();

	PatternFileUI *pfui;
	PatternFile *pf;
	PatternFileMap::iterator i;

	if ((i = patternFileMap.find (path)) != patternFileMap.end()) {
		(*i).second->show_all ();
		return;
	}

	try { 
		pf = new PatternFile (*_sequencer, path);
	} 

	catch (failed_constructor& err) {
		error << "Could not open pattern file \""
		      << path
		      << '"'
		      << endmsg;
		return;
	}

	pfui = new PatternFileUI (*pf, *_sequencer);

	std::pair<Glib::ustring, PatternFileUI *> v;
	
	v.first = path;
	v.second = pfui;
	
	patternFileMap.insert (v);

	pfui->show_all ();
	return;
}

void
SequencerUI::open_new_pattern_file ()
{
	PatternFileUI *pfui;
	PatternFile *pf;

	try {
		pf = new PatternFile (*_sequencer);
	}
	catch (failed_constructor& err) {
		error << "Could not open new pattern file"
		      << endmsg;
		return;
	}

	pfui = new PatternFileUI (*pf, *_sequencer);
	pfui->show_all ();
}


void
SequencerUI::sync ()
{
	_sequencer->sync ();
}


int
SequencerUI::grab_port_name (const MIDI::Port &port, size_t n, void *arg)
{
	if (strlen (port.name()) > strlen (longest_port_string)) {
		longest_port_string = port.name ();
	}

	SequencerUI::_port_strings.push_back (const_cast<char*> (port.name()));
	return 0;
}

void
SequencerUI::get_descriptive_strings ()
{
	int n;

	MIDI::Manager::instance()->foreach_port_tag (grab_port_name, 0);

	for (n = 0; MIDI::controller_names[n]; n++) {
		const char *str;

		str = MIDI::controller_names[n];

		if (strlen (longest_controller_string) < strlen (str)) {
			longest_controller_string = str;
		}

		_controller_strings.push_back ((char *) str);
	}

	clock_mode_strings.push_back (_("Internal"));
	clock_mode_strings.push_back (_("External"));
}

void
integer_filter (char buf[32], Gtk::Adjustment &adj, void *arg)
{
	sprintf (buf, "%d", (int) adj.get_value());
}

void
SequencerUI::clock_filter (char buf[32], Gtk::Adjustment &adj, void *arg)
{
	SequencerUI *gui = (SequencerUI *) arg;

	if (gui->clock_mode == external) {
		strcpy (buf, "External");
	} else {
		sprintf (buf, "%zd BPM: %zd ticks/beat", 
			 (size_t) adj.get_value(),
			 Config->ticks_per_beat);
	}
}

int 
SequencerUI::set_sequencer (Sequencer& seq)
{
	_sequencer = &seq;

	if (setup_windows ()) {
		return -1;
	}

	if (setup_tick_pipe ()){ 
		return -1;
	}

	/* make the sequencer start when we're ready */

	starting.connect (mem_fun (seq, &Sequencer::start));
	stopping.connect (mem_fun (seq, &Sequencer::halt));

	return 0;
}

void
SequencerUI::add_actions ()
{
	Glib::RefPtr<ActionGroup> main_actions = ActionGroup::create (X_("Main"));
	Glib::RefPtr<Action> act;

	ActionManager::register_action (main_actions, X_("File"), _("File"));
	ActionManager::register_action (main_actions, X_("NewFile"), _("New Pattern File"), mem_fun (*this, &SequencerUI::open_new_pattern_file));
	ActionManager::register_action (main_actions, X_("OpenFile"), _("Open Pattern File"), mem_fun (*this, &SequencerUI::open_pattern_file));
	ActionManager::register_action (main_actions, X_("AddSequence"), _("Add Sequence"), mem_fun (*this, &SequencerUI::new_sequence));
	ActionManager::register_action (main_actions, X_("Exit"), _("Exit"), mem_fun (*this, &SequencerUI::quit));

	Glib::RefPtr<ActionGroup> transport_actions = ActionGroup::create (X_("transport"));
	act = ActionManager::register_action (transport_actions, X_("Transport"), _("Transport"));
	ActionManager::transport_actions.push_back (act);
	act = ActionManager::register_action (transport_actions, X_("toggle"), X_("Play"), mem_fun (*this, &SequencerUI::toggle_pause));
	ActionManager::transport_actions.push_back (act);
	act->property_stock_id() = Gtk::Stock::MEDIA_PAUSE;
	act = ActionManager::register_action (transport_actions, X_("rew"), X_("Sync"), mem_fun (*this, &SequencerUI::sync));
	act->property_stock_id() = Gtk::Stock::MEDIA_REWIND;
	ActionManager::transport_actions.push_back (act);

	Glib::RefPtr<ActionGroup> trigger_actions = ActionGroup::create (X_("triggers"));
	ActionManager::register_action (trigger_actions, X_("Triggers"), _("Triggers"));
	ActionManager::register_action (trigger_actions, X_("DummyTriggerEdit"), _("DummyTriggerEdit"));
	ActionManager::register_action (trigger_actions, X_("NewTrigger"), _("New Trigger"), (mem_fun (*this, &SequencerUI::add_trigger)));
	ActionManager::register_action (trigger_actions, X_("EditTrigger"), _("Edit Trigger"));

	ActionManager::add_action_group (main_actions);
	ActionManager::add_action_group (transport_actions);
	ActionManager::add_action_group (trigger_actions);
}
