/*
    Copyright (C) 1998-2007 Paul Davis
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: volume_controller.cc,v 1.4 2000/05/03 15:54:21 pbd Exp $
*/

#include <string.h>
#include <limits.h>

#include <midi++/port.h>
#include <midi++/channel.h>

#include "ui.h"
#include "volume_controller.h"
#include "sequence.h"

using namespace Gtk;

VolumeController::VolumeController (Glib::RefPtr<Gdk::Pixbuf> p,
				    Sequence &seq,
				    Gtk::Adjustment *adj,
				    bool with_numeric) 

	: MIDI::Controllable (seq.read_port(), seq.input_channel()),
	  MotionFeedback (p, MotionFeedback::Rotary, "", adj, false),
	  sequence (seq),
	  adjustment (adj),
	  prompter (Gtk::WIN_POS_MOUSE)
{			  
	new_value_pending = false;

	adjustment->signal_value_changed().connect(mem_fun (*this,&VolumeController::do_adjustment));

	seq.input_midi_changed.connect (mem_fun (*this, &VolumeController::reset_midi));
	learning_started.connect (mem_fun (*this, &VolumeController::prompt));
	learning_stopped.connect (mem_fun (*this, &VolumeController::unprompt));
}

void
VolumeController::do_adjustment ()
{
	sequence.set_volume ((int) floor (adjustment->get_value()));
}

gint
VolumeController::button_press (GdkEventButton *ev)

{
        /* XXX make this configurable */

	if ((ev->state & GDK_CONTROL_MASK) && ev->button == 2) { 
		learn_about_external_control ();
		return TRUE;
	}
	return FALSE;
}

gint
learning_timeout (void *arg)

{
	((VolumeController *) arg)->timed_out ();
	return FALSE;
}

void
VolumeController::timed_out ()

{
	unprompt ();
	stop_learning ();
}

void
VolumeController::prompt ()

{
	prompter.set_text ("Operate MIDI controller now");
	prompter.touch ();
	timeout_tag = gtk_timeout_add (30000, learning_timeout, this);
}

void
VolumeController::unprompt ()
{
	gtk_timeout_remove (timeout_tag);
	prompter.remove ();
}

gint
VolumeController::update_controller_value (void *arg)
{
	VolumeController *c = (VolumeController *) arg;

	c->adjustment->set_value (c->new_value);
	c->new_value_pending = false;

	/* must return false - intended to be used as a oneshot idle callback */
	return false;
}

void
VolumeController::set_value (float v)
{
	/* This is called from a MIDI callback. It could happen
	   a thousand times a second, or more. Therefore, instead
	   of going straight to the X server, which won't work for
	   thread-related reasons, simply request an update whenever
	   the GTK main loop is idle.
	*/
	
	new_value = v;
	sequence.set_volume ((int) v);

	if (!new_value_pending) {
		new_value_pending = true;
		SequencerUI::instance()->idle_add (update_controller_value,this);
	}
}

void
VolumeController::reset_midi ()
{
	midi_rebind (sequence.read_port(), sequence.input_channel());
}
